/*
======================================================================
about.c

Ernie Wright  06 Feb 07

Display an About dialog, including the version and build date.

   aboutDialog()

8 Jun 07    Bumped the version to 1.1.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <gtk/gtk.h>
#include "astro.h"

#define AW 284
#define AH 125


/* only destroy the dialog if it exists */

static int exists;


/*
======================================================================
dismiss()
timercb()

Callbacks that close the About window.  dismiss() is called for key
and mouse button events, and timercb() is called for GLib event loop
timeouts.
====================================================================== */

static gint dismiss( GtkWidget *window, GdkEvent *event, gpointer aboutdlg )
{
   if ( exists ) {
      exists = 0;
      gtk_widget_destroy( GTK_WIDGET( aboutdlg ));
   }
   return TRUE;
}


static gboolean timercb( gpointer aboutdlg )
{
   if ( exists ) {
      exists = 0;
      gtk_widget_destroy( GTK_WIDGET( aboutdlg ));
   }
   return FALSE;
}


/*
======================================================================
create_image()

Create the logo image displayed in the About window.  The base image
is loaded from a file, and the version and build date are written on
it.  It's then returned as a GtkImage.
====================================================================== */

static GtkWidget *create_image( GtkWidget *window )
{
   static GdkPixmap *pixmap;
   
   if ( !pixmap ) {
      PangoFontDescription *fontdesc;
      PangoLayout *layout;
      char imagefile[ MAX_PATH_LEN ];
      
      if ( findFile( GTK_WINDOW( window ), "about.png", imagefile,
         sizeof( imagefile )))
      {
         pixmap = gdk_pixmap_create_from_xpm( window->window, NULL, NULL,
            imagefile );
         fontdesc = pango_font_description_from_string( "sans bold 12" );
         layout = gtk_widget_create_pango_layout( window, NULL );
         pango_layout_set_font_description( layout, fontdesc );
         pango_layout_set_text( layout, "version 1.1\nbuild " __DATE__, -1 );
         gdk_draw_layout( pixmap, window->style->fg_gc[ 0 ], 14, 84, layout );
      }
   }
   return gtk_image_new_from_pixmap( pixmap, NULL );
}


/*
======================================================================
init()

Create an About dialog, a borderless window containing a logo image.
The user can dismiss the window by clicking on it or by pressing any
key.  The window is modal and centered on its parent.
====================================================================== */

static GtkWidget *create_dialog( GtkWidget *window )
{
   GtkWidget *aboutdlg;
   GtkWidget *image;
   GtkWidget *eventbox;
      
   aboutdlg = gtk_window_new( GTK_WINDOW_TOPLEVEL );
   gtk_window_set_transient_for( GTK_WINDOW( aboutdlg ), GTK_WINDOW( window ));

   image = create_image( window );
   eventbox = gtk_event_box_new();
   gtk_container_add( GTK_CONTAINER( eventbox ), image );
   gtk_container_add( GTK_CONTAINER( aboutdlg ), eventbox );

   GTK_WIDGET_SET_FLAGS( eventbox, GTK_CAN_FOCUS );
   gtk_widget_add_events( eventbox,
      GDK_BUTTON_PRESS_MASK | GDK_KEY_PRESS_MASK );
   g_signal_connect( eventbox, "key-press-event", G_CALLBACK( dismiss ),
      aboutdlg );
   g_signal_connect( eventbox, "button-press-event", G_CALLBACK( dismiss ),
      aboutdlg );

   gtk_window_set_decorated( GTK_WINDOW( aboutdlg ), FALSE );
   gtk_window_set_position( GTK_WINDOW( aboutdlg ),
      GTK_WIN_POS_CENTER_ON_PARENT );
   gtk_window_set_modal( GTK_WINDOW( aboutdlg ), TRUE );

   return aboutdlg;
}


/*
======================================================================
aboutDialog()

Display the About dialog.  The timeout value is in milliseconds.  If
this is greater than 0, the dialog will close itself after the timeout
has elapsed.
====================================================================== */

void aboutDialog( GtkWindow *window, int timeout )
{
   GtkWidget *aboutdlg;
   
   aboutdlg = create_dialog( GTK_WIDGET( window ));
   if ( timeout > 0 )
      g_timeout_add( timeout, timercb, aboutdlg );
   exists = 1;
   gtk_widget_show_all( aboutdlg );
}
