/*
======================================================================
astro.h

Ernie Wright  08 Jul 06

Definitions and prototypes for Almagest Stars.

8 Jun 07    Added support for changing epoch and lat/lon offsets.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

/* astro.h */

#ifndef ASTRO_H
#define ASTRO_H

#include "vecmat.h"
#include "view.h"
#include "ptolemy.h"

/* a few significant Julian days */

#define JD_15OCT1582 2299161  /* Catholic church, 4 Oct followed by 15 Oct */
#define JD_14SEP1752 2361222  /* Britain and colonies, 2 Sep followed by 14 Sep */
#define JD_J2000     2451545  /* epoch of J2000 coordinates */

/* pi, and some unit conversions */

#define PI            3.1415926535897932385
#define TWOPI         (2.0*PI)
#define HALFPI        (0.5*PI)
#define DEGRAD(x)     ((x)*(PI/180.0))            /* degrees to radians */
#define RADDEG(x)     ((x)*(180.0/PI))            /* radians to degrees */
#define HRRAD(x)      ((x)*(PI/12.0))             /* hour angle to radians */
#define RADHR(x)      ((x)*(12.0/PI))             /* radians to hour angle */
#define ASECRAD(x)    ((x)*(PI/(180.0*3600.0)))   /* arcseconds to radians */
#define RADASEC(x)    ((x)*(180.0*3600.0/PI))     /* radians to arcseconds */
#define SEXDEC(x,m,s) ((x)+(m)/60.0+(s)/3600.0)   /* sex to decimal */
#define ROUND(x)      ((x)>0?(int)((x)+0.5):-(int)(0.5-(x)))

/* precess.c */

void   nutation( double t, double *dpsi, double *deps );
double obliq( double jd );
void   eq_ecl( double jd, Matrix3 M );
void   ecl_eq( double jd, Matrix3 M );
void   precess_ecl( double jd, Matrix3 M );
void   precess2_ecl( double from, double to, Matrix3 M );

/* atime.c */

double cal_jd( int month, int day, int year );
double caltm_jd( int month, int day, int year, int hour, int min, int sec );
void   jd_cal( double jd, int *month, int *day, int *year );
void   jd_caltm( double jd, int *month, int *day, int *year, int *hour,
         int *min, int *sec );
int    jd_dow( double jd );
int    jd_dpm( double jd );
double jd_year( double jd );

/* yale.c */

typedef struct st_YaleStar {
   int      hr;               /* Harvard/BSC catalog number */
   int      hd;               /* Henry Draper catalog number */
   char     name[ 16 ];       /* Bayer/Flamsteed designation */
   float    ra;               /* right ascension, radians */
   float    dec;              /* declination, J2000, radians */
   float    dist;             /* distance in parsecs */
   float    pmra;             /* proper motion in right ascension, asec/yr */
   float    pmdec;            /* proper motion in declination, asec/yr */
   float    rvel;             /* radial velocity, km/sec */
   float    mag;              /* visual magnitude */
   float    bvindex;          /* B-V index */
   int      cindex;           /* color index */
   Vector   pos;              /* cartesian coordinates */
   float    x;                /* display x coordinate */
   float    y;                /* display y coordinate */
   int      visible;          /* whether currently visible on the display */
} YaleStar;

void posYale( YaleStar star[], double jd );
int  readYale( char *filename, YaleStar star[] );
int  clickYale( YaleStar star[], int x, int y, double pradius, int found[] );
int  HRtoIndex( int hr );

/* color table index from B-V, in colortab.c */

int indexFromBV( double bv );

/* colors are stored as COLORREFs (like Win32's, but backwards) */

typedef unsigned int COLORREF;
#define RGB(r,g,b)     ((COLORREF)((r)<<16|(g)<<8|(b)))
#define RGB16(r,g,b)   ((COLORREF)(((r)/257)<<16|((g)/257)<<8|((b)/257)))
#define GetRValue(c)   (((c)>>16)&0xFF)
#define GetGValue(c)   (((c)>>8)&0xFF)
#define GetBValue(c)   ((c)&0xFF)
#define GetRValue16(c) (257*GetRValue(c))
#define GetGValue16(c) (257*GetGValue(c))
#define GetBValue16(c) (257*GetBValue(c))
#define NO_COLOR       ((COLORREF) -1)

/* arrow key modes */

#define ARROW_MODE_VIEW 0
#define ARROW_MODE_PTOLEMY 1

/* element visibility */

#define LVIS_YALE 1
#define LVIS_PTOLEMY 2
#define LVIS_PLABEL 4
#define LVIS_IDLINE 8
#define LVIS_COORDLN 16
#define LVIS_RADECLN 32
#define LVIS_STARS (LVIS_YALE | LVIS_PTOLEMY)

/* drawbuf.c, drawpdf.c */

typedef struct st_DrawInfo {
   YaleStar             *star;
   PtolemyConstellation *pcon;
   float                 errLon;
   float                 errLat;
   View                 *view;
   int                   layervis;
   double                jd;
   char                 *catname;
   char                 *filename;
} DrawInfo;

/* drawi.c */

typedef struct {
   double x, y;
} ArcPoint;

typedef void ( *PolyLineFunc )( ArcPoint arcpt[], int pcount );
typedef void ( *DrawLineFunc )( double x0, double y0, double x1, double y1 );

void drawTransform( PtolemyConstellation *pcon, YaleStar star[], View *view,
   float dlon, float dlat );
void getLabelPos( PtolemyStar *pstar, YaleStar star[], int layervis,
   double tw, double th, double pradius, double *x, double *y );
int setStatusText( char *buf, View *view, char *pconname,
   float errLon, float errLat, char *catname, int arrowmode );
void drawIDLines( PtolemyConstellation *pcon, YaleStar *star, double mx,
   double my, DrawLineFunc drawline );
void drawCoordLines( View *view, PolyLineFunc plfunc );
void drawRADecLines( View *view, double jd, PolyLineFunc plfunc );

/* drawpdf.c */

int drawPDF( DrawInfo *di );

#ifdef __GTK_H__

/* drawbuf.c */

void initDraw( void );
void sizeDBuf( int w, int h );
void drawDBuf( DrawInfo *di, GdkDrawable *drawable, GdkGC *gc,
   PangoLayout *layout );

COLORREF getPrefColor( int i );
void setPrefColor( int i, COLORREF c );

#define PREFCOLOR_LABEL 2
#define PREFCOLOR_GRID 4
#define PREFCOLOR_RADEC 5
#define PREFCOLOR_RADECTXT 6

void setPangoColor( PangoLayout *layout, COLORREF fg, COLORREF bg );

/* menu.c */

void menuSetCheck( int id, int state );
void menuSetLabel( int id, char *label );
GtkWidget *createMenu( GtkWidget *window, GCallback menu_proc,
   GCallback menu_open );

/* files.c */

#define MAX_PATH_LEN 512

int fileRequest( GtkWindow *window, char *filename, char *buf, int bufsize,
   int type, char *caption );
int findFile( GtkWindow *window, char *filename, char *buf, int bufsize );
int selectColor( GtkWindow *window, COLORREF *color );

#define FREQ_LOAD 1
#define FREQ_SAVE 2

/* colortab.c */

void initColorTable( COLORREF ctab[], double saturation );

/* info.c */

int clickInfo( GtkWindow *window, YaleStar star[], PtolemyConstellation *pcon,
   int x, int y, double pradius, float offsetLon, float offsetLat );

/* epoch.c */

int dlgEpoch( GtkWindow *window, double *jd, float *offsetLon, float *offsetLat );
double getDefaultEpoch( void );

/* about.c */

void aboutDialog( GtkWindow *window, int timeout );

/* config.c */

int loadConfig( char *filename, int *width, int *height, ConViewDefaults cv[],
   char *helpcmd[], double *epoch, float *offsetLon, float *offsetLat );
int saveConfig( char *filename, int width, int height, ConViewDefaults cv[],
   char *helpcmd[], double epoch, float offsetLon, float offsetLat );

#endif

#endif
