/*
======================================================================
atime.c

Ernie Wright  8 Jul 06

Functions for converting between Julian and calendar dates.

   cal_jd()       date to Julian day
   caltm_jd()     date and time to Julian day
   jd_cal()       Julian day to date
   jd_caltm()     Julian day to date and time
   jd_dow()       day of the week
   jd_dpm()       days in the month
   jd_year()      Julian day to fractional year
   year_jd()      fractional year to Julian day

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <math.h>
#include "astro.h"


/*
======================================================================
cal_jd()

Convert a UT calendar date to a Julian day number.

   month    calendar month, 1 to 12
   day      calendar day, 1 to 31
   year     calendar year

Returns the Julian date at noon on the given calendar day.  The start
of the day, at midnight, is the return value minus 0.5.
====================================================================== */

double cal_jd( int month, int day, int year )
{
   long j, m, y, q;

   if ( year < 0 ) ++year;
   if ( month > 2 ) { y = year;     m = month +  1; }
   else             { y = year - 1; m = month + 13; }

   q = y / 4 - (( y < 0 ) && ( y % 4 ));
   j = 365 * y + q + ( 306 * m ) / 10 + day + 1720995;
   if ( j >= 2299161 ) j += 2 - ( y / 100 ) + ( y / 400 );
   return ( double ) j;
}


/*
======================================================================
caltm_jd()

Convert a UT calendar date and time to a Julian date.

   month    calendar month, 1 to 12
   day      calendar day, 1 to 31
   year     calendar year
   hour     hours since midnight, 0 to 23
   min      minutes since the start of the hour, 0 to 59
   sec      seconds since the start of the minute, 0 to 59

Returns the Julian date at noon on the given calendar day.  The start
of the day, at midnight, is the return value minus 0.5.
====================================================================== */

double caltm_jd( int month, int day, int year, int hour, int min, int sec )
{
   return cal_jd( month, day, year ) - 0.5
      + hour / 24.0 + min / 1440.0 + sec / 86400.0;
}


/*
======================================================================
jd_cal()

Convert a Julian date to a UT calendar date.

   jd       Julian date
   month    storage for calendar month, 1 to 12
   day      storage for calendar day, 1 to 31
   year     storage for calendar year
====================================================================== */

void jd_cal( double jd, int *month, int *day, int *year )
{
   long a, b, c, d, e, g, j;

   j = ( long )( jd + 0.5 );
   if ( j >= 2299161 ) {
      g = ( 4 * ( j - 1867216 ) - 1 ) / 146097;
      a = j + 1 + g - ( g / 4 );
   }
   else a = j;

   b = a + 1524;
   c = ( 97594800 + 40 * ( b - 2439870 ) - 4884 ) / 14610;
   d = 365 * c + c / 4;
   e = ( 100 * ( b - d )) / 3061;

   *day   = ( int )( b - d - ( 306 * e ) / 10 );
   *month = ( int )( e - 1 );
   *year  = ( int )( c - 4715 );

   if ( *month > 12 ) *month -= 12;
   if ( *month > 2 ) --( *year );
   if ( *year <= 0 ) --( *year );
}


/*
======================================================================
jd_caltm()

Convert a Julian date to a UT calendar date and time.

   jd       Julian date
   month    storage for calendar month, 1 to 12
   day      storage for calendar day, 1 to 31
   year     storage for calendar year
   hour     storage for the hour, 0 to 23
   min      storage for the minute, 0 to 59
   sec      storage for the second, 0 to 59
====================================================================== */

void jd_caltm( double jd, int *month, int *day, int *year, int *hour,
   int *min, int *sec )
{
   int s;

   jd_cal( jd, month, day, year );
   jd -= 0.5;
   s = ( int )( 86400 * ( jd - floor( jd )));
   *hour = s / 1440;
   *min = ( s % 1440 ) / 60;
   *sec = s % 60;
}


/*
======================================================================
jd_dow()

Day of the week (Sunday == 0) for a julian date.
====================================================================== */

int jd_dow( double jd )
{
   int dow;
   dow = ( int )((( long ) floor( jd - 0.5 ) + 2 ) % 7 );
   if ( dow < 0 ) dow += 7;
   return dow;
}


/*
======================================================================
jd_dpm()

Number of days in the month in which the jd lies, accounting for the
difference between Gregorian and Julian Februarys and returning 21 for
Oct 1582 as a special case.
====================================================================== */

int jd_dpm( double jd )
{
   static short dpm[] = {
      31, 28, 31, 30,
      31, 30, 31, 31,
      30, 31, 30, 31
   };
   int m, d, y;

   jd_cal( jd, &m, &d, &y );

   switch ( m ) {
      case 2:  return ( y > 1582 ) ?
                  28 + (( !( y % 4 ) && y % 100 ) || !( y % 400 )) :
                  28 + !( y % 4 );
      case 10: return ( y == 1582 ) ? 21 : 31;
      default: return dpm[ m - 1 ];
   }
}


/*
======================================================================
jd_year()

Convert a Julian date to a year with a fractional part.  For negative
years, the fractional part is the *positive* offset from the start of
the year.  -1.2 is in February of 1 BC, not November of 2 BC.
====================================================================== */

double jd_year( double jd )
{
   int m, d, y;
   jd_cal( jd, &m, &d, &y );
   return y + ( jd - cal_jd( 1, 1, y )) / 365.25;
}


/*
======================================================================
year_jd()

Convert a year with a fractional part to a Julian date.
====================================================================== */

double year_jd( double year )
{
   int y;
   y = ( int ) year;
   return cal_jd( 1, 1, y ) + fabs( year - y ) * 365.25;
}
