/*
======================================================================
cats.c

Ernie Wright  20 Sep 06

Basic Ptolemy catalog functions.

   readPCat()              allocate and read a catalog
   createPCat()            allocate a PtolemyCatalog
   freePCat()              free a PtolemyCatalog
   freePCon()              free constellation memory
   freePCons()             for all constellations
   posPtolemyCon()         initialize position fields
   clickPtolemy()          find star nearest to mouse click
   getConViewDefaults()    viewing defaults for a constellation
   initConViewDefaults()   viewing defaults for all constellations

5 Jun 07    Significant rewrite to support the PtolemyCatalog data
            structure and dynamic allocation and freeing of catalogs.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "astro.h"


/*
======================================================================
posPtolemyCon()

Initialize spherical and vector position fields.

The Ptolemaic coordinates

   longitude:
      zodiac constellation (0 - 11)
      degrees from the start of the constellation (0 - 29)
      minutes within the degree (0, 10, 15, 20, 30, 40, 45, 50)
   latitude:
      sign (0 = north, 1 = south)
      degrees from the ecliptic
      minutes within the degree

are converted to lon, lat in radians and a 3D vector.  Offsets in
longitude and latitude (radians) are added to lon, lat and included in
the vector.
====================================================================== */

void posPtolemyCon( PtolemyConstellation *pcon, float offsetLon,
   float offsetLat )
{
   int j;
   
   for ( j = 0; j < pcon->nstars; j++ ) {
      pcon->star[ j ].lat = ( float ) DEGRAD(
         pcon->star[ j ].latDeg + pcon->star[ j ].latMin / 60.0 );
      if ( pcon->star[ j ].latNeg )
         pcon->star[ j ].lat *= -1;
      pcon->star[ j ].lat += offsetLat;
      pcon->star[ j ].lon = ( float ) DEGRAD(
         pcon->star[ j ].lonZod * 30 + pcon->star[ j ].lonDeg
         + pcon->star[ j ].lonMin / 60.0 );
      pcon->star[ j ].lon += offsetLon;
      sphere_vec( pcon->star[ j ].lon, pcon->star[ j ].lat,
         pcon->star[ j ].pos );
   }
}


/*
======================================================================
clickPtolemy()

Find the star within pradius of (x, y).

Given a constellation, an (x, y) position and a distance, returns the
index of the star with x, y fields closer than the distance.  If more
than one star is within the distance, the closest is returned.  If no
star is within the distance, -1 is returned.

Used to identify the Ptolemy star closest to a mouse click, where the
(x, y) positions and the distance are in screen pixels.
====================================================================== */

int clickPtolemy( PtolemyConstellation *pcon, int x, int y, double pradius )
{
   double dx, dy, d, dmin = 10000.0;
   int imin = -1, i;

   for ( i = 0; i < pcon->inside + pcon->outside; i++ ) {
      if ( !pcon->star[ i ].visible ) continue;
      dx = x - pcon->star[ i ].x;
      dy = y - pcon->star[ i ].y;
      d = sqrt( dx * dx + dy * dy );
      if ( d < dmin && d <= pradius ) {
         dmin = d;
         imin = i;
      }
   }
   return imin;
}


/*
----------------------------------------------------------------------
Default viewing parameters for each Ptolemaic constellation.

   heading
   pitch
   bank
   zoom
   longitude shift
   latitude shift
---------------------------------------------------------------------- */

static float defview[ NPCONS ][ 6 ] = {
    80.0f,  73.0f, 0.0f,  8.0f, 0.8f, -0.3f,
   116.0f,  38.0f, 0.0f,  4.5f, 1.1f,  0.4f,
   115.0f,  79.0f, 0.0f,  4.5f, 0.5f,  0.0f,
     6.0f,  68.0f, 0.0f,  6.0f, 1.3f, -0.2f,
   173.0f,  43.0f, 0.0f,  4.5f, 1.2f,  0.6f,
   196.0f,  47.0f, 0.0f,  8.0f, 1.2f,  0.3f,
   217.0f,  55.0f, 0.0f,  4.5f, 1.8f,  0.3f,
   266.0f,  59.0f, 0.0f,  8.0f, 2.2f,  0.0f,
   293.0f,  59.0f, 0.0f,  6.0f, 1.0f,  0.1f,
    17.0f,  49.0f, 0.0f,  8.0f, 1.5f, -0.3f,
    36.0f,  25.0f, 0.0f,  6.0f, 1.4f, -0.3f,
    56.0f,  18.0f, 0.0f,  6.0f, 1.0f,  0.2f,
   230.0f,  17.0f, 0.0f,  4.5f, 1.2f,  0.2f,
   227.0f,  27.0f, 0.0f,  4.5f, 1.7f,  0.2f,
   276.0f,  39.0f, 0.0f,  8.0f, 1.1f,  0.2f,
   270.0f,  26.0f, 0.0f,  8.0f, 1.7f,  0.4f,
   291.0f,  31.0f, 0.0f,  8.0f, 1.5f,  0.3f,
   298.0f,  23.0f, 0.0f, 10.0f, 1.3f,  0.3f,
   325.0f,  27.0f, 0.0f,  6.0f, 1.3f,  0.0f,
     1.0f,  30.0f, 0.0f,  6.0f, 1.6f, -0.4f,
    14.0f,  19.0f, 0.0f,  8.0f, 0.4f, -0.2f,
    17.0f,   3.0f, 0.0f,  8.0f, 0.9f, -0.6f,
    44.0f,  -5.0f, 0.0f,  6.0f, 1.1f,  0.2f,
    78.0f,   0.0f, 0.0f,  6.0f, 1.0f,  0.0f,
   101.0f,   1.0f, 0.0f,  8.0f, 1.3f, -0.6f,
   128.0f,  10.0f, 0.0f,  4.5f, 1.4f,  0.2f,
   170.0f,   5.0f, 0.0f,  6.0f, 1.3f,  0.2f,
   206.0f,   2.0f, 0.0f,  6.0f, 1.2f,  0.0f,
   228.0f,  -8.0f, 0.0f,  6.0f, 1.2f, -0.5f,
   258.0f,  -9.0f, 0.0f,  6.0f, 1.3f, -0.7f,
   287.0f,   0.0f, 0.0f,  8.0f, 0.9f, -0.2f,
   307.0f,  -3.0f, 0.0f,  4.5f, 0.9f, -0.4f,
   344.0f,   6.0f, 0.0f,  6.0f, 1.2f, -0.4f,
   356.0f, -18.0f, 0.0f,  6.0f, 0.6f, -1.2f,
    55.0f, -19.0f, 0.0f,  6.0f, 0.7f, -0.4f,
    26.0f, -41.0f, 0.0f,  6.0f, 0.3f, -1.1f,
    54.0f, -40.0f, 0.0f,  8.0f, 0.2f,  0.0f,
    74.0f, -45.0f, 0.0f,  4.5f, 0.8f, -0.9f,
    88.0f, -15.0f, 0.0f, 10.0f, 1.2f,  0.0f,
   113.0f, -61.0f, 0.0f,  6.0f, 1.0f, -0.5f,
   146.0f, -22.0f, 0.0f,  2.5f, 1.2f,  0.0f,
   152.0f, -17.0f, 0.0f,  8.0f, 1.1f, -0.1f,
   167.0f, -16.0f, 0.0f,  8.0f, 1.3f,  0.0f,
   198.0f, -36.0f, 0.0f,  4.5f, 1.5f,  0.3f,
   212.0f, -22.0f, 0.0f,  8.0f, 1.5f,  0.0f,
   238.0f, -27.0f, 0.0f,  8.0f, 3.0f,  0.8f,
   253.0f, -17.0f, 0.0f,  8.0f, 1.0f,  0.0f,
   293.0f, -20.0f, 0.0f,  8.0f, 1.0f, -1.0f
};


/*
----------------------------------------------------------------------
Name and size of each constellation.
---------------------------------------------------------------------- */

static struct {
   int   inside;             /* number of stars in the figure */
   int   outside;            /* number of stars outside the figure */
   char *name;               /* constellation name */
} defcon[ NPCONS ] =
{
    7,  1, "Ursa Minor",
   27,  8, "Ursa Major",
   31,  0, "Draco",
   11,  2, "Cepheus",
   22,  1, "Bootes",
    8,  0, "Corona Borealis",
   29,  1, "Hercules",
   10,  0, "Lyra",
   17,  2, "Cygnus",
   13,  0, "Cassiopeia",
   26,  3, "Perseus",
   14,  0, "Auriga",
   24,  5, "Ophiuchus",
   18,  0, "Serpens",
    5,  0, "Sagitta",
    9,  6, "Aquila",
   10,  0, "Delphinus",
    4,  0, "Equuleus",
   20,  0, "Pegasus",
   23,  0, "Andromeda",
    4,  0, "Triangulum",
   13,  5, "Aries",
   33, 11, "Taurus",
   18,  7, "Gemini",
    9,  4, "Cancer",
   27,  8, "Leo",
   26,  6, "Virgo",
    8,  9, "Libra",
   21,  3, "Scorpius",
   31,  0, "Sagittarius",
   28,  0, "Capricornus",
   42,  3, "Aquarius",
   34,  4, "Pisces",
   22,  0, "Cetus",
   38,  0, "Orion",
   34,  0, "Eridanus",
   12,  0, "Lepus",
   18, 11, "Canis Major",
    2,  0, "Canis Minor",
   45,  0, "Argo Navis",
   25,  2, "Hydra",
    7,  0, "Crater",
    7,  0, "Corvus",
   37,  0, "Centaurus",
   19,  0, "Lupus",
    7,  0, "Ara",
   13,  0, "Corona Australis",
   12,  6, "Piscis Austrinus"
};


/*
======================================================================
getConViewDefaults()

Set default viewing parameters for a constellation.
====================================================================== */

void getConViewDefaults( ConViewDefaults *cv, int i )
{
   cv->lon    = ( float ) DEGRAD( defview[ i ][ 0 ] );
   cv->lat    = ( float ) DEGRAD( defview[ i ][ 1 ] );
   cv->bank   = ( float ) DEGRAD( defview[ i ][ 2 ] );
   cv->zoom   = defview[ i ][ 3 ];
   cv->errLon = ( float ) DEGRAD( defview[ i ][ 4 ] );
   cv->errLat = ( float ) DEGRAD( defview[ i ][ 5 ] );
}


/*
======================================================================
initConViewDefaults()

Set default viewing parameters for all constellations.
====================================================================== */

void initConViewDefaults( ConViewDefaults cv[] )
{
   int i;
   for ( i = 0; i < NPCONS; i++ )
      getConViewDefaults( &cv[ i ], i );
}


/*
======================================================================
freePCon()

Free allocated memory used by a PtolemyConstellation:  the name of the
constellation, the description string for each star, and the array of
stars.  Does not free the PtolemyConstellation itself.
====================================================================== */

void freePCon( PtolemyConstellation *pcon )
{
   int i;
   
   if ( pcon ) {
      if ( pcon->name )
         free( pcon->name );
      if ( pcon->star ) {
         for ( i = 0; i < pcon->nstars; i++ )
            if ( pcon->star[ i ].description )
               free( pcon->star[ i ].description );
         free( pcon->star );
      }
   }
}


/*
======================================================================
freePCons()

Free memory used within an array of PtolemyConstellations.  Does not
free the array itself.
====================================================================== */

void freePCons( PtolemyConstellation pcon[] )
{
   int i;
   
   for ( i = 0; i < NPCONS; i++ )
      freePCon( &pcon[ i ] );
}


/*
======================================================================
freePCat()

Free memory allocated by createPCat().  Returns NULL as a convenience,
for constructions like

   if ( error ) return freePCat( pcat );
   
which is used in createPCat() and readPCat().
====================================================================== */

PtolemyCatalog *freePCat( PtolemyCatalog *pcat )
{
   if ( pcat ) {
      freePCons( pcat->con );
      if ( pcat->filename )
         free( pcat->filename );
      if ( pcat->name )
         free( pcat->name );
      free( pcat );
   }
   return NULL;
}


/*
======================================================================
createPCat()

Allocate and initialize a PtolemyCatalog.  The filename is the name of
the file containing the star data for the catalog.  The catname is the
descriptive name of the catalog.  Either or both may be NULL or empty.
====================================================================== */

PtolemyCatalog *createPCat( char *filename, char *catname )
{
   PtolemyCatalog *pcat;
   int i;
      
   pcat = calloc( 1, sizeof( PtolemyCatalog ));
   if ( !pcat ) return NULL;
   
   if ( filename && filename[ 0 ] ) {
      pcat->filename = malloc( strlen( filename ) + 1 );
      if ( !pcat->filename )
         return freePCat( pcat );
      strcpy( pcat->filename, filename );
   }
   
   if ( catname && catname[ 0 ] ) {
      pcat->name = malloc( strlen( catname ) + 1 );
      if ( !pcat->name )
         return freePCat( pcat );
      strcpy( pcat->name, catname );
   }
   
   for ( i = 0; i < NPCONS; i++ ) {
      pcat->con[ i ].name = malloc( strlen( defcon[ i ].name ) + 1 );
      if ( !pcat->con[ i ].name )
         return freePCat( pcat );
      strcpy( pcat->con[ i ].name, defcon[ i ].name );
      pcat->con[ i ].inside = defcon[ i ].inside;
      pcat->con[ i ].outside = defcon[ i ].outside;
      pcat->con[ i ].nstars = defcon[ i ].inside + defcon[ i ].outside;
      pcat->con[ i ].star = calloc( pcat->con[ i ].nstars,
         sizeof( PtolemyStar ));
      if ( !pcat->con[ i ].star )
         return freePCat( pcat );
   }

   return pcat;
}
   

/*
======================================================================
getline()

Read one line from a text file.  Initializes the buffer to all zeroes,
so that code reading from areas of the buffer beyond EOL don't read
leftover garbage.  Removes trailing carriage return and line feed
bytes.
====================================================================== */

static int getline( FILE *fp, char *buf, int bufsize )
{
   char *p;

   memset( buf, 0, bufsize );
   if ( !fgets( buf, bufsize, fp ))
      return 0;
   if ( p = strrchr( buf, 0x0D ))
      *p = 0;
   if ( p = strrchr( buf, 0x0A ))
      *p = 0;
   return 1;
}


/*
======================================================================
readPCat()

Read a Ptolemy star catalog file.  Allocates and initializes a
PtolemyCatalog and fills its star arrays with data from the named
file.  Returns the PtolemyCatalog if successful, otherwise NULL.
====================================================================== */

#define TOKEN(o, s)   (memcpy( tok, &buf[ o ], s ), tok[ s ] = '\0')
#define TOKENI(o, s)  (TOKEN(o, s), atoi( tok ))
#define TOKENF(o, s)  (TOKEN(o, s), ( float ) atof( tok ))

PtolemyCatalog *readPCat( char *filename )
{
   PtolemyCatalog *pcat;
   PtolemyStar *star;
   char buf[ 512 ], tok[ 80 ];
   int i, j, nstars;
   FILE *fp;

   fp = fopen( filename, "r" );
   if ( !fp )
      return 0;

   if ( !getline( fp, buf, sizeof( buf ))) {
      fclose( fp );
      return 0;
   }

   if ( strncmp( buf, "Almagest Catalog:", 17 )) {
      fclose( fp );
      return 0;
   }

   pcat = createPCat( filename, buf + 19 );
   if ( !pcat ) return NULL;

   for ( i = 0; i < NPCONS; i++ ) {
      for ( j = 0; j < pcat->con[ i ].nstars; j++ ) {
         if ( !getline( fp, buf, sizeof( buf )))
            return freePCat( pcat );
         star = &pcat->con[ i ].star[ j ];

         star->baily       = TOKENI( 0, 4 );
         star->index       = TOKENI( 9, 2 );
         star->lonDeg      = TOKENI( 12, 3 );
         star->lonZod      = star->lonDeg / 30;
         star->lonDeg     %= 30;
         star->lonMin      = TOKENI( 16, 2 );
         star->latNeg      = ( buf[ 19 ] == '-' );
         star->latDeg      = TOKENI( 20, 2 );
         star->latMin      = TOKENI( 23, 2 );
         star->mag         = TOKENF( 26, 3 );
         star->yaleIndex   = TOKENI( 30, 4 );
         star->description = malloc( strlen( buf + 48 ) + 1 );
         strcpy( star->description, buf + 48 );

         TOKEN( 35, 12 );
         strcpy( star->id, tok );
      }
   }

   fclose( fp );
   return pcat;
}
