/*
======================================================================
config.c

Ernie Wright  20 Sep 06

Load and save program state.

   saveConfig()
   loadConfig()

8 Jun 07    Added epoch and lat/lon offsets.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include "astro.h"


/*
----------------------------------------------------------------------
Data in the config file is saved as a set of name-value pairs.  The
value can comprise multiple components and span multiple lines.  The
enum values are used to index the tag strings.
---------------------------------------------------------------------- */

#define NTAGS 8

enum Tag {
   CFGTAG_NULL,
   CFGTAG_WINDOWRECT,
   CFGTAG_GRIDCOLOR,
   CFGTAG_LABELCOLOR,
   CFGTAG_EQGRIDCOLOR,
   CFGTAG_HELPCOMMAND,
   CFGTAG_CONVIEWARRAY,
   CFGTAG_EPOCH,
   CFGTAG_POFFSETS
};

static struct {
   char      *name;
   enum Tag   tag;
} taglist[ NTAGS ] =
{
   "WindowRect",     CFGTAG_WINDOWRECT,
   "GridColor",      CFGTAG_GRIDCOLOR,
   "LabelColor",     CFGTAG_LABELCOLOR,
   "EqGridColor",    CFGTAG_EQGRIDCOLOR,
   "HelpCommand",    CFGTAG_HELPCOMMAND,
   "ConViewArray",   CFGTAG_CONVIEWARRAY,
   "Epoch",          CFGTAG_EPOCH,
   "POffsets",       CFGTAG_POFFSETS
};


/*
======================================================================
tag_name()
match_tag()

Convert between enum value and tag string.
====================================================================== */

static char *tag_name( enum Tag tag )
{
   int i;

   for ( i = 0; i < NTAGS; i++ )
      if ( tag == taglist[ i ].tag )
         return taglist[ i ].name;
   return NULL;
}


static enum Tag match_tag( char *buf )
{
   int i;

   for ( i = 0; i < NTAGS; i++ )
      if ( !strcmp( buf, taglist[ i ].name ))
         return taglist[ i ].tag;
   return CFGTAG_NULL;
}


/*
======================================================================
saveConfig()

Write config values to filename.  The width and height are the size of
the display window.  helpcmd[] contains the name of a program and its
first argument, which are used to invoke a program to display the HTML
help file.
====================================================================== */

int saveConfig( char *filename, int width, int height, ConViewDefaults cv[],
   char *helpcmd[], double epoch, float offsetLon, float offsetLat )
{
   FILE *fp;
   COLORREF c;
   int i;

   fp = fopen( filename, "w" );
   if ( !fp ) return 0;

   fprintf( fp, "%s %d %d %d %d\n", tag_name( CFGTAG_WINDOWRECT ),
      0, 0, width, height );

   c = getPrefColor( PREFCOLOR_GRID );
   fprintf( fp, "%s %d %d %d\n", tag_name( CFGTAG_GRIDCOLOR ),
      GetRValue( c ), GetGValue( c ), GetBValue( c ));

   c = getPrefColor( PREFCOLOR_LABEL );
   fprintf( fp, "%s %d %d %d\n", tag_name( CFGTAG_LABELCOLOR ),
      GetRValue( c ), GetGValue( c ), GetBValue( c ));

   c = getPrefColor( PREFCOLOR_RADEC );
   fprintf( fp, "%s %d %d %d\n", tag_name( CFGTAG_EQGRIDCOLOR ),
      GetRValue( c ), GetGValue( c ), GetBValue( c ));

   if ( helpcmd ) {
      fprintf( fp, "%s", tag_name( CFGTAG_HELPCOMMAND ));
      if ( helpcmd[ 0 ] ) {
         fprintf( fp, " %s", helpcmd[ 0 ] );
         if ( helpcmd[ 1 ] )
            fprintf( fp, " %s", helpcmd[ 1 ] );
      }
      fprintf( fp, "\n" );
   }
   
   fprintf( fp, "%s %f\n", tag_name( CFGTAG_EPOCH ), epoch );
   
   fprintf( fp, "%s %f %f\n", tag_name( CFGTAG_POFFSETS ),
      RADDEG( offsetLon ), RADDEG( offsetLat ));
      
   fprintf( fp, "%s\n", tag_name( CFGTAG_CONVIEWARRAY ));
   for ( i = 0; i < 48; i++ )
      fprintf( fp, "   %5.1f %5.1f %5.1f %5.1f %5.1f %5.1f\n",
         RADDEG( cv[ i ].lon ),
         RADDEG( cv[ i ].lat ),
         RADDEG( cv[ i ].bank ),
         cv[ i ].zoom,
         RADDEG( cv[ i ].errLon ),
         RADDEG( cv[ i ].errLat ));

   fclose( fp );
   return 1;
}


/*
======================================================================
init()

Initialize config values with built-in defaults.  These are what
loadConfig() will return for items it doesn't read from the config
file.
====================================================================== */

static void init( int *width, int *height, ConViewDefaults cv[],
   double *epoch, float *offsetLon, float *offsetLat )
{
   GdkRectangle r;
   int i;
   
   setPrefColor( PREFCOLOR_LABEL, RGB( 64, 96, 64 ));
   setPrefColor( PREFCOLOR_GRID, RGB( 32, 32, 64 ));
   setPrefColor( PREFCOLOR_RADEC, RGB( 64, 32, 32 ));
   initConViewDefaults( cv );

   gdk_window_get_frame_extents( gdk_get_default_root_window(), &r );
   i = r.width / 8;
   *width  = r.width  - 2 * i;
   *height = r.height - 2 * i;
   
   *epoch = getDefaultEpoch();
   *offsetLon = *offsetLat = 0.0f;
}


/*
======================================================================
getline()

Read a line from the file.  Returns 0 if feof() is true or fgets()
fails.  Zeroes out the line buffer and removes trailing carriage
return and linefeed characters.
====================================================================== */

static int getline( FILE *fp, char *buf, int bufsize )
{
   char *p;

   if ( feof( fp ))
      return 0;
   memset( buf, 0, bufsize );
   if ( !fgets( buf, bufsize, fp ))
      return 0;
   if ( p = strrchr( buf, 0x0D ))
      *p = 0;
   if ( p = strrchr( buf, 0x0A ))
      *p = 0;
   return 1;
}


/*
======================================================================
getint()
getfloat()
getangle()
getstring()

Extract a value from a string.  We're using strtok().  Since in most
cases strtok() has already been called on the line buffer to get the
tag name, the buf argument will be NULL, but a non-NULL buf can be
passed for lines that don't begin with a tag.
====================================================================== */

static int getint( char *buf )
{
   return atoi( strtok( buf, " " ));
}


static float getfloat( char *buf )
{
   return ( float ) atof( strtok( buf, " " ));
}


static double getdouble( char *buf )
{
   return atof( strtok( buf, " " ));
}


static float getangle( char *buf )
{
   return ( float ) DEGRAD( atof( strtok( buf, " " )));
}


static char *getstring( char *buf )
{
   char *p, *s = NULL;
   
   p = strtok( buf, " " );
   if ( p ) {
      s = malloc( strlen( p ) + 1 );
      strcpy( s, p );
   }
   return s;
}


/*
======================================================================
loadConfig()

Read config values from a file written by saveConfig.  If the file
can't be opened or its contents are incomplete, hard-coded defaults
are supplied.
====================================================================== */

int loadConfig( char *filename, int *width, int *height, ConViewDefaults cv[],
   char *helpcmd[], double *epoch, float *offsetLon, float *offsetLat )
{
   FILE *fp;
   enum Tag tag;
   int i, red, green, blue, x, y;
   char buf[ 128 ], *tok;

   init( width, height, cv, epoch, offsetLon, offsetLat );
   if ( !findFile( NULL, ".almstars", filename, MAX_PATH_LEN ))
      return 0;

   fp = fopen( filename, "r" );
   if ( !fp ) return 0;

   while ( 1 ) {
      if ( !getline( fp, buf, sizeof( buf )))
         break;
      tok = strtok( buf, " " );
      tag = match_tag( tok );
      if ( !tag ) continue;

      switch ( tag ) {
         case CFGTAG_WINDOWRECT:
            x = getint( NULL );
            y = getint( NULL );
            *width  = getint( NULL ) - x;
            *height = getint( NULL ) - y;
            break;
         case CFGTAG_GRIDCOLOR:
            red   = getint( NULL );
            green = getint( NULL );
            blue  = getint( NULL );
            setPrefColor( PREFCOLOR_GRID, RGB( red, green, blue ));
            break;
         case CFGTAG_LABELCOLOR:
            red   = getint( NULL );
            green = getint( NULL );
            blue  = getint( NULL );
            setPrefColor( PREFCOLOR_LABEL, RGB( red, green, blue ));
            break;
         case CFGTAG_EQGRIDCOLOR:
            red   = getint( NULL );
            green = getint( NULL );
            blue  = getint( NULL );
            setPrefColor( PREFCOLOR_RADEC, RGB( red, green, blue ));
            break;
         case CFGTAG_HELPCOMMAND:
            helpcmd[ 0 ] = getstring( NULL );
            helpcmd[ 1 ] = getstring( NULL );
            break;
         case CFGTAG_EPOCH:
            *epoch = getdouble( NULL );
            break;
         case CFGTAG_POFFSETS:
            *offsetLon = getangle( NULL );
            *offsetLat = getangle( NULL );
            break;
         case CFGTAG_CONVIEWARRAY:
            for ( i = 0; i < 48; i++ ) {
               getline( fp, buf, sizeof( buf ));
               cv[ i ].lon    = getangle( buf );
               cv[ i ].lat    = getangle( NULL );
               cv[ i ].bank   = getangle( NULL );
               cv[ i ].zoom   = getfloat( NULL );
               cv[ i ].errLon = getangle( NULL );
               cv[ i ].errLat = getangle( NULL );
            }
            break;
         default:
            break;
      }
   }

   fclose( fp );
   return 1;
}
