/*
======================================================================
drawbuf.c

Ernie Wright  29 Jul 06, 10 Jan 07

Drawing routines for the display buffer, an array of 3-byte RGB pixels
sized to the client area of the main window.

   initDraw()           call once to initialize the star color table
   sizeDBuf()           whenever you get a window resize event
   getPrefColor()       return user-defined color
   setPrefColor()       store user-defined color
   setPangoColor()      set text color
   drawDBuf()           draw the display

The display rendering is grafted onto a local but nearly independent
set of low-level raster drawing routines that support antialiasing,
subpixel resolution and some compositing.

   clearDBuf()          erase the buffer, convert it to the local
   realizeDBuf()        representation of a displayable raster

   scaleColor()         create a color by scaling an input color or
   addColors()          by combining two inputs in various ways
   combineColors()
   meanColors()

   getPixelDBuf()       basic pixel get and set
   setPixelDBuf()

   addPixelDBuf()       set pixel based on an input color and the
   combinePixelDBuf()   existing buffer color

   drawDot()            drawing primitives:  dot, circle, line
   drawCircle()
   drawLine()

   inDot()              support for dots and circles, which are drawn
   onCircle()           by testing pixels in a bounding box

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gtk/gtk.h>
#include "astro.h"


/* parameters and memory for our display buffer */

static COLORREF             ctab[ 191 ];
static COLORREF             prefcolor[ 7 ];
static unsigned char       *bits;
static int                  width, height;
static int                  rowbytes;


/*
======================================================================
initDraw()

Any one-time-only initialization happens here.
====================================================================== */

void initDraw( void )
{
   initColorTable( ctab, 1.0 );
}


/*
======================================================================
sizeDBuf()

Change the display buffer when the window size changes.
====================================================================== */

void sizeDBuf( int w, int h )
{
   if ( bits ) free( bits );

   width = w;
   height = h;
   rowbytes = (( width * 3 + 3 ) >> 2 ) << 2;
   bits = malloc( rowbytes * height );
}


/*
======================================================================
setPangoColor()

Set the foreground and background color attributes of a PangoLayout.
If the background color is NO_COLOR, the text background will be
transparent.
====================================================================== */

void setPangoColor( PangoLayout *layout, COLORREF fg, COLORREF bg )
{
   PangoAttribute *fgattr;
   PangoAttrList *alist;

   fgattr = pango_attr_foreground_new(
      GetRValue16( fg ),
      GetGValue16( fg ),
      GetBValue16( fg ));
   fgattr->start_index = 0;
   fgattr->end_index = G_MAXUINT;
   
   alist = pango_attr_list_new();
   pango_attr_list_insert( alist, fgattr );
   
   if ( bg != NO_COLOR ) {
      PangoAttribute *bgattr = pango_attr_background_new(
         GetRValue16( bg ),
         GetGValue16( bg ),
         GetBValue16( bg ));
      bgattr->start_index = 0;
      bgattr->end_index = G_MAXUINT;
      pango_attr_list_insert( alist, bgattr );
   }
      
   pango_layout_set_attributes( layout, alist );
}

   
/*
======================================================================
clearDBuf()

Fill the display buffer image with black.
====================================================================== */

static void clearDBuf( void )
{
   memset( bits, 0, rowbytes * height );
}


/*
======================================================================
realizeDBuf()

Copy the display buffer to a GDK drawable.
====================================================================== */

static void realizeDBuf( GdkDrawable *drawable, GdkGC *gc )
{
   gdk_draw_rgb_image( drawable, gc, 0, 0, width, height,
      GDK_RGB_DITHER_NORMAL, bits, rowbytes );
}


/*
======================================================================
scaleColor()

Scale the levels in the COLORREF by a factor s.
====================================================================== */

static COLORREF scaleColor( COLORREF color, double s )
{
   int r, g, b;

   r = ( int )( s * GetRValue( color ) + 0.5 );
   g = ( int )( s * GetGValue( color ) + 0.5 );
   b = ( int )( s * GetBValue( color ) + 0.5 );
   if ( r > 255 ) r = 255;  if ( r < 0 ) r = 0;
   if ( g > 255 ) g = 255;  if ( g < 0 ) g = 0;
   if ( b > 255 ) b = 255;  if ( b < 0 ) b = 0;
   return RGB( r, g, b );
}


/*
======================================================================
addColors()

Add two COLORREFs.
====================================================================== */

static COLORREF addColors( COLORREF c1, COLORREF c2 )
{
   int r, g, b;

   r = GetRValue( c1 ) + GetRValue( c2 );
   g = GetGValue( c1 ) + GetGValue( c2 );
   b = GetBValue( c1 ) + GetBValue( c2 );
   if ( r > 255 ) r = 255;  if ( r < 0 ) r = 0;
   if ( g > 255 ) g = 255;  if ( g < 0 ) g = 0;
   if ( b > 255 ) b = 255;  if ( b < 0 ) b = 0;
   return RGB( r, g, b );
}


/*
======================================================================
combineColors()

Combine two COLORREFs.  The combination is defined as the maximum of
the two colors for each RGB component.
====================================================================== */

static COLORREF combineColors( COLORREF c1, COLORREF c2 )
{
   int r, g, b, r1, g1, b1, r2, g2, b2;

   r1 = GetRValue( c1 );
   g1 = GetGValue( c1 );
   b1 = GetBValue( c1 );
   r2 = GetRValue( c2 );
   g2 = GetGValue( c2 );
   b2 = GetBValue( c2 );

   r = r1 > r2 ? r1 : r2;
   g = g1 > g2 ? g1 : g2;
   b = b1 > b2 ? b1 : b2;

   if ( r > 255 ) r = 255;  if ( r < 0 ) r = 0;
   if ( g > 255 ) g = 255;  if ( g < 0 ) g = 0;
   if ( b > 255 ) b = 255;  if ( b < 0 ) b = 0;
   return RGB( r, g, b );
}


/*
======================================================================
meanColors()

Return the mean of two COLORREFs.
====================================================================== */

static COLORREF meanColors( COLORREF c1, COLORREF c2 )
{
   int r, g, b, r1, g1, b1, r2, g2, b2;

   r1 = GetRValue( c1 );
   g1 = GetGValue( c1 );
   b1 = GetBValue( c1 );
   r2 = GetRValue( c2 );
   g2 = GetGValue( c2 );
   b2 = GetBValue( c2 );

   r = ( r1 + r2 ) >> 1;
   g = ( g1 + g2 ) >> 1;
   b = ( b1 + b2 ) >> 1;
   return RGB( r, g, b );
}


/*
======================================================================
getPrefColor()

Return a user-defined color.
====================================================================== */

COLORREF getPrefColor( int i )
{
   return prefcolor[ i ];
}


/*
======================================================================
setPrefColor()

Set a user-defined color.
====================================================================== */

void setPrefColor( int i, COLORREF c )
{
   prefcolor[ i ] = c;
   if ( i == PREFCOLOR_RADEC )
      prefcolor[ PREFCOLOR_RADECTXT ] = meanColors( c, RGB( 255, 255, 255 ));
}


/*
======================================================================
getPixelDBuf()

Return the RGB value of the pixel at (x, y).
====================================================================== */

static COLORREF getPixelDBuf( int x, int y )
{
   unsigned char *p;
   if ( x < 0 || x > width - 1 || y < 0 || y > height - 1 ) return 0;
   p = bits + y * rowbytes + x * 3;
   return RGB( p[ 0 ], p[ 1 ], p[ 2 ] );
}


/*
======================================================================
setPixelDBuf()

Set the color of the pixel at (x, y).
====================================================================== */

static void setPixelDBuf( int x, int y, COLORREF c )
{
   unsigned char *p;
   if ( x < 0 || x > width - 1 || y < 0 || y > height - 1 ) return;
   p = bits + y * rowbytes + x * 3;
   p[ 0 ] = GetRValue( c );
   p[ 1 ] = GetGValue( c );
   p[ 2 ] = GetBValue( c );
}


/*
======================================================================
addPixelDBuf()

Scale the input color by factor p, then add it with the existing RGB
value at (x, y) and set the pixel.
====================================================================== */

static void addPixelDBuf( int x, int y, COLORREF c, double p )
{
   setPixelDBuf( x, y, addColors( getPixelDBuf( x, y ), scaleColor( c, p )));
}


/*
======================================================================
combinePixelDBuf()

Scale the input color by factor p, then combine it with the existing
RGB value at (x, y) and set the pixel.
====================================================================== */

static void combinePixelDBuf( int x, int y, COLORREF c, double p )
{
   setPixelDBuf( x, y, combineColors( getPixelDBuf( x, y ),
      scaleColor( c, p )));
}


/*
======================================================================
inDot()

Return the intensity of a pixel for a dot of a given radius.

   ix, iy   pixel to test
   x, y     center of the dot
   r        pixel radius of the dot

If the distance from (x, y) to (ix, iy) is less than r - 0.5, the
intensity is 1.0.  The intensity falls off linearly over distances
ranging from r - 0.5 to r + 0.5, beyond which it's 0.0.
====================================================================== */

static double inDot( int ix, int iy, double x, double y, double r )
{
   double dx, dy, dr;

   dx = x - ( ix + 0.5 );
   dy = y - ( iy + 0.5 );
   dr = sqrt( dx * dx + dy * dy ) - r;

   if ( dr < -0.5 ) return 1.0;
   if ( dr >  0.5 ) return 0.0;
   return 0.5 - dr;
}


/*
======================================================================
onCircle()

If (ix, iy) is near the circle centered at (x, y) with radius r,
returns a pixel intensity between 0 and 1.  Otherwise returns 0.
====================================================================== */

static double onCircle( int ix, int iy, double x, double y, double r )
{
   double dx, dy, dr;

   dx = x - ( ix + 0.5 );
   dy = y - ( iy + 0.5 );
   if ( fabs( dx ) + fabs( dy ) < r - 1 ) return 0.0;
   dr = fabs( sqrt( dx * dx + dy * dy ) - r );

   if ( dr < 1.6 )
      return 0.625 * ( 1.6 - dr );
   else
      return 0.0;
}


/*
======================================================================
drawDot()

Draws a dot centered at (x, y) with radius r.  A rectangle of pixels
centered on (x, y) is tested.
====================================================================== */

static void drawDot( double x, double y, double r, COLORREF color,
   void ( *setpixel )( int x, int y, COLORREF c, double p ))
{
   double p;
   int ix, iy, ir, j, k;

   ir = ( int ) ceil( r ) + 1;
   for ( j = -ir; j <= ir; j++ ) {
      iy = ( int )( y + 0.5 ) + j;
      if ( iy < 0 || iy >= height ) continue;

      for ( k = -ir; k <= ir; k++ ) {
         ix = ( int )( x + 0.5f ) + k;
         if ( ix < 0 || ix >= width ) continue;

         p = inDot( ix, iy, x, y, r );
         if ( p > 0.0 )
            setpixel( ix, iy, color, p );
      }
   }
}


/*
======================================================================
drawCircle()

Draws a circle centered at (x, y) with radius r.  The circle is
antialiased, and both its center and its radius will in general be
non-integral.  We test every pixel within a square that encloses the
circle.

Called by drawPtolemyStar().
====================================================================== */

static void drawCircle( double x, double y, double r, COLORREF color )
{
   double p;
   int ix, iy, ir, j, k;

   ir = ( int ) ceil( r ) + 1;
   for ( j = -ir; j <= ir; j++ ) {
      iy = ( int )( y + 0.5 ) + j;
      if ( iy < 0 || iy >= height ) continue;

      for ( k = -ir; k <= ir; k++ ) {
         ix = ( int )( x + 0.5f ) + k;
         if ( ix < 0 || ix >= width ) continue;

         p = onCircle( ix, iy, x, y, r );
         if ( p > 0.0 )
            combinePixelDBuf( ix, iy, color, p );
      }
   }
}


/*
======================================================================
drawLine()

Draw a line from (fx0, fy0) to (fx1, fy1).  The line is antialiased,
and the endpoints will in general be non-integral.  Bresenham is used
to pick the pixels closest to the line, then that pixel and two of its
neighbors are colored according to their distance from the true line.

Called by drawIDLine().
====================================================================== */

static void drawLine( double fx0, double fy0, double fx1, double fy1,
   COLORREF color, void ( *setpixel )( int x, int y, COLORREF c, double p ))
{
   int x0, y0, x1, y1, d, x, y, ax, ay, sx, sy, dx, dy;
   double fdx, fdy, fdx2, fdy2, len, p;

   fdx = fx1 - fx0;
   fdy = fy1 - fy0;
   len = sqrt( fdx * fdx + fdy * fdy );
   if ( len < 0.5 ) return;

   x0 = ( int )( fx0 + 0.5 );
   y0 = ( int )( fy0 + 0.5 );
   x1 = ( int )( fx1 + 0.5 );
   y1 = ( int )( fy1 + 0.5 );

   dy = y1 - y0;
   dx = x1 - x0;
   ax = abs( dx ) << 1;
   ay = abs( dy ) << 1;
   sx = dx < 0 ? -1 : 1;
   sy = dy < 0 ? -1 : 1;

   if ( x0 < 0 ) x0 = 0;  if ( x0 > width  - 1 ) x0 = width  - 1;
   if ( y0 < 0 ) y0 = 0;  if ( y0 > height - 1 ) y0 = height - 1;
   if ( x1 < 0 ) x1 = 0;  if ( x1 > width  - 1 ) x1 = width  - 1;
   if ( y1 < 0 ) y1 = 0;  if ( y1 > height - 1 ) y1 = height - 1;

   x = x0;
   y = y0;

   if ( ax > ay ) {                    /* x dominant */
      d = ay - ( ax >> 1 );
      for ( ;; ) {
         fdx2 = fx0 - x;
         fdy2 = fy0 - y;

         /* draw the central pixel */
         p = ( 1.5 - fabs( fdx * fdy2 - fdy * fdx2 ) / len ) / 1.5;
         if ( p > 0.0 )
            setpixel( x, y, color, p );

         /* draw the one below it */
         if ( y < height - 1 ) {
            p = ( 1.5 - fabs( fdx * ( fdy2 - 1 ) - fdy * fdx2 ) / len ) / 1.5;
            if ( p > 0.0 )
               setpixel( x, y + 1, color, p );
         }

         /* draw the one above it */
         if ( y > 0 ) {
            p = ( 1.5 - fabs( fdx * ( fdy2 + 1 ) - fdy * fdx2 ) / len ) / 1.5;
            if ( p > 0.0 )
               setpixel( x, y - 1, color, p );
         }

         if ( x == x1 ) return;
         if ( d >= 0 ) {
            y += sy;
            d -= ax;
         }
         x += sx;
         d += ay;
      }
   }
   else {                              /* y dominant */
      d = ax - ( ay >> 1 );
      for ( ;; ) {
         fdx2 = fx0 - x;
         fdy2 = fy0 - y;

         /* draw the central pixel */
         p = ( 1.5 - fabs( fdx * fdy2 - fdy * fdx2 ) / len ) / 1.5;
         if ( p > 0.0 )
            setpixel( x, y, color, p );

         /* draw the one to the right of it */
         if ( x < width - 1 ) {
            p = ( 1.5 - fabs( fdx * fdy2 - fdy * ( fdx2 - 1 )) / len ) / 1.5;
            if ( p > 0.0 )
               setpixel( x + 1, y, color, p );
         }

         /* draw the one to the left of it */
         if ( x > 0 ) {
            p = ( 1.5 - fabs( fdx * fdy2 - fdy * ( fdx2 + 1 )) / len ) / 1.5;
            if ( p > 0.0 )
               setpixel( x - 1, y, color, p );
         }

         if ( y == y1 ) return;
         if ( d >= 0 ) {
            x += sx;
            d -= ay;
         }
         y += sy;
         d += ax;
      }
   }
}


/*
======================================================================
draw_stardot()

Draw a dot representing a star.  The color, brightness, and radius of
the dot reflect the color and magnitude of the star.
====================================================================== */

static void draw_stardot( float x, float y, float mag, COLORREF cindex )
{
   double r, s;
   int c;

   if ( mag > 5.0 ) {
      r = 0.5;
      s = 1.25 / ( mag - 3.75 );
      c = scaleColor( ctab[ cindex ], s );
      drawDot( x, y, r, c, addPixelDBuf );
   }
   else {
      r = 0.5 + 0.5 * ( 5.0 - mag );
      drawDot( x, y, r, ctab[ cindex ], addPixelDBuf );
   }
}


/*
======================================================================
draw_yalestar()

Draw a Yale star using draw_stardot().
====================================================================== */

static void draw_yalestar( YaleStar *star )
{
   draw_stardot( star->x, star->y, star->mag, star->cindex );
}


/*
======================================================================
draw_ptolemystar()

Draw a Ptolemy star using either draw_circle() or draw_stardot().
====================================================================== */

static void draw_ptolemystar( PtolemyStar *star, int layervis, double r )
{
   COLORREF c;

   if ( layervis & LVIS_PTOLEMY ) {
      if ( star->mag > 6.5f )  /* "nebulous" */
         c = RGB( 36, 48, 80 );
      else
         c = scaleColor( RGB( 160, 160, 160 ), ( 8.0 - star->mag ) / 6.5 );
      drawCircle( star->x, star->y, r, c );
   }
   else
      draw_stardot( star->x, star->y, star->mag, 0 );
}


/*
======================================================================
draw_labels()

Write numbers identifying Almagest stars.
====================================================================== */

static void draw_labels( GdkDrawable *drawable, GdkGC *gc, PangoLayout *layout,
   PtolemyConstellation *pcon, YaleStar star[], int layervis, double pradius )
{
   int i, len, w, h;
   double x, y;
   char buf[ 8 ];

   if ( !( layervis & LVIS_PLABEL )) return;

   setPangoColor( layout, prefcolor[ PREFCOLOR_LABEL ], NO_COLOR );

   for ( i = 0; i < pcon->inside + pcon->outside; i++ ) {
      if ( pcon->star[ i ].visible ) {
         len = sprintf( buf, "%d", pcon->star[ i ].index );
         pango_layout_set_text( layout, buf, len );
         pango_layout_get_pixel_size( layout, &w, &h );
         getLabelPos( &pcon->star[ i ], star, layervis, w, h, pradius, &x, &y );
         y -= h;
         gdk_draw_layout( drawable, gc, ROUND( x ), ROUND( y ), layout );
      }
   }
}


/*
======================================================================
idline_cb()

Callback called by drawIDLines() (in drawi.c) to render an ID line.
====================================================================== */

static void idline_cb( double x0, double y0, double x1, double y1 )
{
   drawLine( x0, y0, x1, y1, RGB( 64, 96, 64 ), combinePixelDBuf );
}


/*
======================================================================
polyline_cb()

Callback used by drawCoordLines() and drawRADecLines() (in drawi.c) to
render a polyline.
====================================================================== */

static COLORREF plcolor;

static void polyline_cb( ArcPoint arcpt[], int pcount )
{
   int i;
   
   if ( pcount < 2 ) return;
   for ( i = 0; i < pcount - 1; i++ )
      drawLine( arcpt[ i ].x, arcpt[ i ].y, arcpt[ i + 1 ].x, arcpt[ i + 1 ].y,
         plcolor, combinePixelDBuf );
}


/*
======================================================================
drawDBuf()

Render the current view.
====================================================================== */

void drawDBuf( DrawInfo *di, GdkDrawable *drawable, GdkGC *gc,
   PangoLayout *layout )
{
   View view;
   int i;

   drawTransform( di->pcon, di->star, di->view, di->errLon, di->errLat );

   clearDBuf();

   view = *di->view;
   --view.width;        /* routines in drawi.c use these as maxx and maxy */
   --view.height;

   plcolor = prefcolor[ PREFCOLOR_GRID ];
   if ( di->layervis & LVIS_COORDLN )
      drawCoordLines( &view, polyline_cb );

   plcolor = prefcolor[ PREFCOLOR_RADEC ];
   if ( di->layervis & LVIS_RADECLN )
      drawRADecLines( &view, di->jd, polyline_cb );

   if ( di->layervis & LVIS_IDLINE )
      drawIDLines( di->pcon, di->star, width - 1, height - 1, idline_cb );

   if (( di->layervis & LVIS_STARS ) != LVIS_YALE )
      for ( i = 0; i < di->pcon->inside + di->pcon->outside; i++ )
         if ( di->pcon->star[ i ].visible )
            draw_ptolemystar( &di->pcon->star[ i ], di->layervis,
               di->view->pradius );

   if ( di->layervis & LVIS_YALE )
      for ( i = 0; i < 9096; i++ )
         if ( di->star[ i ].visible )
            draw_yalestar( &di->star[ i ] );

   realizeDBuf( drawable, gc );
   draw_labels( drawable, gc, layout, di->pcon, di->star, di->layervis,
      di->view->pradius );
}
