/*
======================================================================
drawi.c

Ernie Wright  09 Feb 07

Some drawing functions that are independent of the image type (raster
or vector) and the drawing technology (GTK+, Cairo).

   drawTransform()   ecliptic to image transformation
   getLabelPos()     calculate position of Ptolemy star labels
   setStatusText()   write status line to a string
   drawIDLines()     lines connecting Ptolemy and Yale stars
   drawCoordLines()  ecliptic coordinate grid
   drawRADecLines()  equatorial coordinate grid
   
The last three take as arguments the image dimensions, for clipping,
and a callback that does the actual rendering.  For rasters, maxx and
maxy typically need to be one less than the width and height.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <math.h>
#include "astro.h"


static double maxx, maxy;
static PolyLineFunc polyline;


/*
======================================================================
drawTransform()

Transform star coordinates to display coordinates.
====================================================================== */

void drawTransform( PtolemyConstellation *pcon, YaleStar star[], View *view,
   float dlon, float dlat )
{
   int i;
   Vector pos;

   for ( i = 0; i < 9096; i++ ) {
      transform4( star[ i ].pos, view->Mv, pos );
      star[ i ].visible = viewToPixel( view, pos, &star[ i ].x, &star[ i ].y );
   }

   posPtolemyCon( pcon, dlon, dlat );
   for ( i = 0; i < pcon->inside + pcon->outside; i++ ) {
      transform4( pcon->star[ i ].pos, view->Mv, pos );
      pcon->star[ i ].visible = viewToPixel( view, pos,
         &pcon->star[ i ].x, &pcon->star[ i ].y );
   }
}


/*
======================================================================
getLabelPos()

Calculate the display coordinates of the numerical label for a
Ptolemy star.  If Ptolemy stars are being drawn as circles, the label
is placed below and to the right of the circle.  Otherwise, if the
star is identified with a Yale star, the label goes opposite the ID
line connecting the Ptolemy and Yale star positions.  If there's no
associated Yale star, the label is centered over the Ptolemy star
position.
                 _               X
               /   \            /
              |     |          /             X
               \ _ / X        /

Assumes that the vertical display coordinate for text is relative to
the baseline of the text.  tw and th are the width and height of the
label text.
====================================================================== */

void getLabelPos( PtolemyStar *pstar, YaleStar star[], int layervis,
   double tw, double th, double pradius, double *x, double *y )
{
   int hr, index;
   double dx, dy, d;
   
   if ( layervis & LVIS_PTOLEMY ) {
      *x = pstar->x + 0.7 * pradius;
      *y = pstar->y + 0.7 * pradius + th;
   }
   else {
      hr = pstar->yaleIndex;
      if ( hr ) {
         index = HRtoIndex( hr );
         dx = pstar->x - star[ index ].x;
         dy = pstar->y - star[ index ].y;
         d = sqrt( dx * dx + dy * dy );
         if ( d > 0.0 ) {
            dx /= d;
            dy /= d;
         }
         else
            dy = 1.0;
         *x = pstar->x + tw * dx * 0.7 - tw / 2;
         *y = pstar->y + th * dy * 0.7 + th / 2;
      }
      else {
         *x = pstar->x - tw / 2;
         *y = pstar->y + th / 2;
      }
   }
}


/*
======================================================================
setStatusText()

Fills in the line of text containing the viewing parameters and the
current constellation and catalog.
====================================================================== */

int setStatusText( char *buf, View *view, char *pconname,
   float errLon, float errLat, char *catname, int arrowmode )
{
   double vfov, hfov;
   
   vfov = RADDEG( 2.0 * atan2( 2.0, view->zoom ));
   hfov = RADDEG( 2.0 * atan2( 2.0 * view->width / view->height, view->zoom ));

   return sprintf( buf, "%s  %cView (%d, %d, %d, %.1f)  FOV (%.1f, %.1f)"
      "  %cShift (%.1f, %.1f)  %s",
      pconname,
      arrowmode == ARROW_MODE_VIEW ? '+' : ' ',
      ROUND( RADDEG( view->lon )),
      ROUND( RADDEG( view->lat )),
      ROUND( RADDEG( view->bank )),
      view->zoom,
      hfov, vfov,
      arrowmode == ARROW_MODE_PTOLEMY ? '+' : ' ',
      ROUND( RADDEG( errLon ) * 10 ) / 10.0,
      ROUND( RADDEG( errLat ) * 10 ) / 10.0,
      catname );
}


/*
======================================================================
clip_line()

Clip a line against the image boundaries.
====================================================================== */

static void clip_line( double x0, double y0, double dx, double dy,
   double *x1, double *y1 )
{
   if ( fabs( dx ) < 0.0001 ) {
      *x1 = x0;
      *y1 = dy > 0 ? maxy : 0;
      return;
   }
   if ( fabs( dy ) < 0.0001 ) {
      *x1 = dx > 0 ? maxx : 0;
      *y1 = y0;
      return;
   }
   *x1 = dx > 0 ? maxx : 0;
   *y1 = y0 + ( *x1 - x0 ) * dy / dx;
   if ( *y1 >= 0 && *y1 <= maxy )
      return;

   *y1 = dy > 0 ? maxy : 0;
   *x1 = x0 + ( *y1 - y0 ) * dx / dy;
}


/*
======================================================================
clip_idline()

Calculate the image coordinates for the YaleStar endpoint of a line
between a PtolemyStar and a YaleStar.  If the PtolemyStar isn't
visible, returns 0.  Otherwise x1, y1 get either the coordinates of
the YaleStar, or coordinates clipped against an image boundary.
====================================================================== */

static void clip_idline( PtolemyStar *pstar, YaleStar *ystar, double *x1,
   double *y1 )
{
   double dx, dy;

   if ( !pstar->visible ) return;

   if ( ystar->visible ) {
      *x1 = ystar->x;
      *y1 = ystar->y;
   }
   else {
      dx = ystar->x - pstar->x;
      dy = ystar->y - pstar->y;
      clip_line( pstar->x, pstar->y, dx, dy, x1, y1 );
   }
}


/*
======================================================================
drawIDLines()

Draws a line between each Ptolemy star and the Yale star with which
it's identified.  Calls the drawline() callback to actually render the
line.
====================================================================== */

void drawIDLines( PtolemyConstellation *pcon, YaleStar star[], double mx,
   double my, DrawLineFunc drawline )
{
   int i, id;
   double x1, y1;
   
   maxx = mx;
   maxy = my;

   for ( i = 0; i < pcon->inside + pcon->outside; i++ ) {
      if ( !pcon->star[ i ].visible )
         continue;
      id = HRtoIndex( pcon->star[ i ].yaleIndex );
      if ( id >= 0 ) {
         clip_idline( &pcon->star[ i ], &star[ id ], &x1, &y1 );
         drawline( pcon->star[ i ].x, pcon->star[ i ].y, x1, y1 );
      }
   }
}


/*
----------------------------------------------------------------------
Coordinate gridlines are represented as a connected series of 1-degree
line segments.  The transformed positions of the segment vertices are
written in carc[] by the coordinate-specific drawing functions (e.g.
drawLonArc()).  drawCArc() clips the vertices against the boundaries
of the image and writes the resulting clipped polyline vertices in
arcpt[], passing this to polyline() for rendering.
---------------------------------------------------------------------- */

static struct {
   double    pos[ 3 ];
   ArcPoint  pt;
   int       visible;
} carc[ 360 ];

static ArcPoint arcpt[ 361 ];


/*
======================================================================
draw_carc()

Draw a coordinate arc, a single grid line in either ecliptic or
equatorial coordinates.  Uses the transformed vertex positions in
carc[] to write the vertices of one or more polylines in arcpt[],
which is passed to the polyline() callback for rendering.

The arc is broken into one-degree line segments.  The endpoints of the
segments are

   carc[ 0..n - 1 ], carc[ 1..n % m ]

If carc[] contains all 360 points of a circle, m = n = 360, and the
second endpoint of the last segment will be carc[ 0 ].  Otherwise,
typically m = n + 1, and the second endpoint of the last segment is
carc[ n ].  n is therefore the number of segments, and m is a modulus
used to index the last vertex.

The arc is clipped against the image boundaries.  For each segment:

   if both endpoints are visible
      if this is the first vertex
         add the first endpoint to the vertex list
      add the second endpoint

   if only the second endpoint is visible
      the arc is emerging from an image boundary
      clip the segment against the boundary (move the first endpoint)
      add the first and second endpoints

   if only the first endpoint is visible
      the arc has hit an image boundary
      clip the segment against the boundary (move the second endpoint)
      add the second endpoint
      emit and empty the polyline

   else if neither endpoint is visible
      if the polyline isn't empty
         emit and empty the polyline
         
Basically, skip segments that aren't visible, start a polyline at the
segment where the arc emerges from an image boundary, and finish the
polyline when the arc hits a second boundary.  It's possible for an
arc to become multiple polylines.

In some extreme views, it's possible for the non-visible vertex to be
behind the projection plane (its z-depth is < 0).  The clipping we do
won't work for those, so we just skip the segment.
====================================================================== */

static void draw_carc( int n, int m )
{
   double x0, y0, x1, y1, dx, dy;
   int i, i0, i1, pcount = 0;


   for ( i = 1; i <= n; i++ ) {
      i0 = i - 1;
      i1 = i % m;

      if ( carc[ i0 ].visible && carc[ i1 ].visible ) {
         if ( !pcount ) {
            arcpt[ 0 ] = carc[ i0 ].pt;
            pcount = 1;
         }
         arcpt[ pcount++ ] = carc[ i1 ].pt;
      }
      else if ( carc[ i1 ].visible && carc[ i0 ].pos[ 2 ] > 0.0f ) {
         x1 = carc[ i1 ].pt.x;
         y1 = carc[ i1 ].pt.y;
         dx = carc[ i0 ].pt.x - x1;
         dy = carc[ i0 ].pt.y - y1;
         clip_line( x1, y1, dx, dy, &x0, &y0 );
         if ( !pcount ) {
            arcpt[ 0 ].x = x0;
            arcpt[ 0 ].y = y0;
            pcount = 1;
         }
         arcpt[ pcount++ ] = carc[ i1 ].pt;
      }
      else if ( carc[ i0 ].visible ) {
         if ( carc[ i1 ].pos[ 2 ] <= 0.0f ) {
            if ( pcount ) {
               polyline( arcpt, pcount );
               pcount = 0;
            }
         }
         else {
            x0 = carc[ i0 ].pt.x;
            y0 = carc[ i0 ].pt.y;
            dx = carc[ i1 ].pt.x - x0;
            dy = carc[ i1 ].pt.y - y0;
            clip_line( x0, y0, dx, dy, &x1, &y1 );
            if ( !pcount ) {
               arcpt[ 0 ] = carc[ i0 ].pt;
               pcount = 1;
            }
            arcpt[ pcount ].x = x1;
            arcpt[ pcount ].y = y1;
            ++pcount;
            polyline( arcpt, pcount );
            pcount = 0;
         }
      }
      else if ( pcount ) {
         polyline( arcpt, pcount );
         pcount = 0;
      }
   }

   if ( pcount )
      polyline( arcpt, pcount );
}


/*
======================================================================
draw_lonarc()

Initialize the carc[] array for a line of constant longitude.  For
each degree of latitude along the line, the position is converted to
image coordinates.  Points that are visible are connected with line
segments in draw_carc().  ilon is in degrees.

Lines at 90 degree multiples are complete to the poles.  Those at 30
degree increments stop 1 degree from the poles, and the rest stop at
15 degrees.  This keeps the pole from getting crowded with converging
longitude lines.
====================================================================== */

static void draw_lonarc( View *view, int ilon )
{
   double lon, lat, pos[ 3 ];
   float x, y;
   int i, n, lo, vcount = 0;

   if (( ilon % 90 ) == 0 ) {
      n = 180;
      lo = 0;
   }
   else if (( ilon % 30 ) == 0 ) {
      n = 178;
      lo = 1;
   }
   else {
      n = 150;
      lo = 15;
   }

   lon = DEGRAD( ilon );
   for ( i = 0; i <= n; i++ ) {
      lat = DEGRAD( 90 - ( i + lo ));
      sphere_vec( lon, lat, pos );
      transform4( pos, view->Mv, carc[ i ].pos );
      carc[ i ].visible = viewToPixel( view, carc[ i ].pos, &x, &y );
      carc[ i ].pt.x = x;
      carc[ i ].pt.y = y;
      vcount += carc[ i ].visible;
   }
   if ( vcount )
      draw_carc( n, n + 1 );
}


/*
======================================================================
draw_raarc()

Like draw_lonarc(), but for right ascensions.  ira is in degrees.  jd
is the epoch, used to account for precession.
====================================================================== */

static void draw_raarc( View *view, int ira, double jd )
{
   Matrix3 M;
   double ra, dec, pos[ 3 ];
   float x, y;
   int i, n, lo, vcount = 0;

   if (( ira % 90 ) == 0 ) {
      n = 180;
      lo = 0;
   }
   else if (( ira % 30 ) == 0 ) {
      n = 178;
      lo = 1;
   }
   else {
      n = 150;
      lo = 15;
   }

   identity3( M );
   eq_ecl( jd, M );
   ra = DEGRAD( ira );
   for ( i = 0; i <= n; i++ ) {
      dec = DEGRAD( 90 - ( i + lo ));
      sphere_vec( ra, dec, pos );
      transform( pos, M, pos );
      transform4( pos, view->Mv, carc[ i ].pos );
      carc[ i ].visible = viewToPixel( view, carc[ i ].pos, &x, &y );
      carc[ i ].pt.x = x;
      carc[ i ].pt.y = y;
      vcount += carc[ i ].visible;
   }
   if ( vcount )
      draw_carc( n, n + 1 );
}


/*
======================================================================
draw_latarc()

Initialize the carc[] array for a line of constant latitude.  For each
degree of longitude along the line, the position is converted to image
coordinates.  Points that are visible are connected with line segments
in draw_carc().  ilat is in degrees.
====================================================================== */

static void draw_latarc( View *view, int ilat )
{
   double lon, lat, pos[ 3 ];
   float x, y;
   int i, vcount = 0;

   lat = DEGRAD( ilat );
   for ( i = 0; i < 360; i++ ) {
      lon = DEGRAD( i );
      sphere_vec( lon, lat, pos );
      transform4( pos, view->Mv, carc[ i ].pos );
      carc[ i ].visible = viewToPixel( view, carc[ i ].pos, &x, &y );
      carc[ i ].pt.x = x;
      carc[ i ].pt.y = y;
      vcount += carc[ i ].visible;
   }
   if ( vcount )
      draw_carc( 360, 360 );
}


/*
======================================================================
draw_decarc()

Like draw_latarc(), but for declinations.
====================================================================== */

static void draw_decarc( View *view, int idec, double jd )
{
   Matrix3 M;
   double ra, dec, pos[ 3 ];
   float x, y;
   int i, vcount = 0;

   identity3( M );
   eq_ecl( jd, M );
   dec = DEGRAD( idec );
   for ( i = 0; i < 360; i++ ) {
      ra = DEGRAD( i );
      sphere_vec( ra, dec, pos );
      transform( pos, M, pos );
      transform4( pos, view->Mv, carc[ i ].pos );
      carc[ i ].visible = viewToPixel( view, carc[ i ].pos, &x, &y );
      carc[ i ].pt.x = x;
      carc[ i ].pt.y = y;
      vcount += carc[ i ].visible;
   }
   if ( vcount )
      draw_carc( 360, 360 );
}


/*
======================================================================
drawCoordLines()

Draw the ecliptic grid.  Calls draw_lonarc() and draw_latarc() at
15-degree intervals.
====================================================================== */

void drawCoordLines( View *view, PolyLineFunc plfunc )
{
   int i;

   polyline = plfunc;
   maxx = view->width;
   maxy = view->height;
   for ( i = 0; i < 360; i += 15 )
      draw_lonarc( view, i );
   for ( i = 15; i < 180; i += 15 )
      draw_latarc( view, 90 - i );
}


/*
======================================================================
drawRADecLines()

Draw the equatorial grid.  Calls draw_raarc() and draw_decarc() at
15-degree intervals.
====================================================================== */

void drawRADecLines( View *view, double jd, PolyLineFunc plfunc )
{
   int i;

   polyline = plfunc;
   maxx = view->width;
   maxy = view->height;
   for ( i = 0; i < 360; i += 15 )
      draw_raarc( view, i, jd );
   for ( i = 15; i < 180; i += 15 )
      draw_decarc( view, 90 - i, jd );
}
