/*
======================================================================
drawpdf.c

Ernie Wright  09 Feb 07

Save a star chart (vector) version of the display as a PDF.

   drawPDF()

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <cairo.h>
#include <cairo-pdf.h>
#include <stdio.h>
#include <math.h>
#include "astro.h"


static cairo_t *cr;
static double width, height;
static double dashes[] = { 1.0, 1.0 };


/*
======================================================================
begin()

Create and initialize a cairo_t for rendering a PDF image.  The w and
h arguments are the size of the image in inches.  The default units
for the cairo_t are points (1/72 inch).
====================================================================== */

static int begin( char *filename, double w, double h )
{
   cairo_surface_t *surface;

   width = w * 72.0;
   height = h * 72.0;
   
   surface = cairo_pdf_surface_create( filename, width, height );
   cr = cairo_create( surface );
   cairo_surface_destroy( surface );

   cairo_select_font_face( cr, "Helvetica", CAIRO_FONT_SLANT_NORMAL,
      CAIRO_FONT_WEIGHT_NORMAL );
   cairo_set_font_size( cr, 8.0 );

   cairo_set_line_width( cr, 0.12 );
   cairo_rectangle( cr, 0, 0, width, height );
   cairo_stroke( cr );
   
   return 1;
}


/*
======================================================================
done()

Finalize the PDF image file.
====================================================================== */

static void done( void )
{
   cairo_show_page( cr );
   cairo_destroy( cr );
   cr = NULL;
}


/*
======================================================================
draw_dot()
====================================================================== */

static void draw_dot( double x, double y, double r )
{
   cairo_arc( cr, x, y, r, 0.0, 2 * PI );
   cairo_fill( cr );
}


/*
======================================================================
draw_circle()
====================================================================== */

static void draw_circle( double x, double y, double r, int dotted,
   double thickness )
{
   if ( dotted )
      cairo_set_dash( cr, dashes, 2, 0.0 );
   cairo_set_line_width( cr, thickness );
   cairo_arc( cr, x, y, r, 0, 2 * PI );
   cairo_stroke( cr );
   if ( dotted )
      cairo_set_dash( cr, dashes, 0, 0.0 );
}


/*
======================================================================
draw_stardot()
====================================================================== */

static void draw_stardot( float x, float y, float mag )
{
   double r;

   r = 4.0 * ( 1.0 - mag * 0.14 );
   if ( r < 0.12 ) r = 0.12;
   draw_dot( x, y, r );
}


/*
======================================================================
draw_yalestar()

Draw a Yale star using draw_stardot().
====================================================================== */

static void draw_yalestar( YaleStar *star )
{
   draw_stardot( star->x, star->y, star->mag );
}


/*
======================================================================
draw_ptolemystar()

Draw a Ptolemy star using either draw_circle() or draw_stardot().
====================================================================== */

static void draw_ptolemystar( PtolemyStar *star, int layervis, double r )
{
   int dotted;
   double thickness;

   if ( layervis & LVIS_PTOLEMY ) {
      if ( star->mag > 6.5f ) {  /* "nebulous" */
         dotted = 1;
         thickness = 0.24;
      }
      else {
         dotted = 0;
         thickness = 0.12 + ( 6.5 - star->mag ) / 6.5;
      }
      draw_circle( star->x, star->y, r, dotted, thickness );
   }
   else
      draw_stardot( star->x, star->y, star->mag );
}


/*
======================================================================
draw_status()

Draw the line of text describing the view.
====================================================================== */

static void draw_status( View *view, char *pconname, float errLon,
   float errLat, char *catname )
{
   char buf[ 160 ];
   cairo_text_extents_t extent;

   setStatusText( buf, view, pconname, errLon, errLat, catname, -1 );
   cairo_text_extents( cr, buf, &extent );
   cairo_move_to( cr, extent.height, height - extent.height );
   cairo_show_text( cr, buf );
}


/*
======================================================================
draw_labels()

Label the Almagest stars with numbers.
====================================================================== */

static void draw_labels( YaleStar star[], PtolemyConstellation *pcon,
   int layervis, double pradius )
{
   cairo_text_extents_t size;
   int i, len;
   double x, y;
   char buf[ 8 ];

   if ( !( layervis & LVIS_PLABEL )) return;

   for ( i = 0; i < pcon->inside + pcon->outside; i++ ) {
      if ( pcon->star[ i ].visible ) {
         len = sprintf( buf, "%d", pcon->star[ i ].index );
         cairo_text_extents( cr, buf, &size );
         getLabelPos( &pcon->star[ i ], star, layervis,
            size.width, size.height, pradius, &x, &y );
         cairo_move_to( cr, x, y );
         cairo_show_text( cr, buf );
      }
   }
}


/*
======================================================================
idline_cb()

Callback called by drawIDLines() (in drawi.c) to render an ID line.
====================================================================== */

static void idline_cb( double x0, double y0, double x1, double y1 )
{
   cairo_move_to( cr, x0, y0 );
   cairo_line_to( cr, x1, y1 );
   cairo_stroke( cr );
}


/*
======================================================================
polyline_cb()

Callback used by drawCoordLines() and drawRADecLines() (in drawi.c) to
render a polyline.
====================================================================== */

static void polyline_cb( ArcPoint arcpt[], int pcount )
{
   int i;
   
   if ( pcount < 2 ) return;
   cairo_move_to( cr, arcpt[ 0 ].x, arcpt[ 0 ].y );
   for ( i = 1; i < pcount; i++ )
      cairo_line_to( cr, arcpt[ i ].x, arcpt[ i ].y );
   cairo_stroke( cr );
}


/*
======================================================================
pdfDraw()

Create a PDF file containing a rendering of the current view.
====================================================================== */

int drawPDF( DrawInfo *di )
{
   View view;
   int i;
   float w, h;

   w = 6.75f;
   h = w * ( float ) di->view->height / di->view->width;

   if ( !begin( di->filename, w, h ))
      return 0;

   view = *di->view;
   view.width = width;
   view.height = height;
   view.pradius = height * DEGRAD( 0.5 ) / ( 2.0 * atan2( 1.0, view.zoom ));

   drawTransform( di->pcon, di->star, &view, di->errLon, di->errLat );

   cairo_set_line_width( cr, 0.24 );
   if ( di->layervis & LVIS_IDLINE )
      drawIDLines( di->pcon, di->star, width, height, idline_cb );

   cairo_set_line_width( cr, 0.12 );
   if ( di->layervis & LVIS_COORDLN )
      drawCoordLines( &view, polyline_cb );

   if ( di->layervis & LVIS_RADECLN ) {
      cairo_set_dash( cr, dashes, 2, 0.0 );
      drawRADecLines( &view, di->jd, polyline_cb );
      cairo_set_dash( cr, dashes, 0, 0.0 );
   }

   if (( di->layervis & LVIS_STARS ) != LVIS_YALE )
      for ( i = 0; i < di->pcon->inside + di->pcon->outside; i++ )
         if ( di->pcon->star[ i ].visible )
            draw_ptolemystar( &di->pcon->star[ i ], di->layervis,
               view.pradius );

   if ( di->layervis & LVIS_YALE )
      for ( i = 0; i < 9096; i++ )
         if ( di->star[ i ].visible )
            draw_yalestar( &di->star[ i ] );

   draw_status( &view, di->pcon->name, di->errLon, di->errLat, di->catname );
   draw_labels( di->star, di->pcon, di->layervis, view.pradius );

   done();
   return 1;
}
