/*
======================================================================
epoch.c

Ernie Wright  5 June 07

Dialog for changing the epoch.

   dlgEpoch()           display the dialog
   getDefaultEpoch()    used by config.c to initialize the epoch

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gtk/gtk.h>
#include "astro.h"


/* GTK widgets for user input */

static GtkWidget *entry_epoch;
static GtkWidget *entry_lon;
static GtkWidget *entry_lat;
static GtkWidget *combo_preset;

/* month strings */

static char *month_str[] = {
   "Jan", "Feb", "Mar", "Apr", "May", "Jun",
   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

/* epoch and lat/lon offset presets */

#define NPRESETS 2

static struct {
   double jd;
   float  lon, lat;
   char  *name;
} preset[ NPRESETS ] =
{
   1771298.0,  0.0f,          0.0f, "Ptolemy",
   1673941.0, -0.0465421134f, 0.0f, "Hipparchus"
};



/*
======================================================================
getDefaultEpoch()

Returns the JD in the first element of the preset array.
====================================================================== */

double getDefaultEpoch( void )
{
   return preset[ 0 ].jd;
}


/*
======================================================================
sprint_date()

Constructs a date string of the form "1 Jan 2000 AD".
====================================================================== */

static char *sprint_date( char *buf, double jd )
{
   int day, month, year;
   
   jd_cal( jd, &month, &day, &year );
   sprintf( buf, "%d %s %d %s", day, month_str[ month - 1 ], abs( year ),
      year < 0 ? "BC" : "AD" );
   return buf;
}


/*
======================================================================
sprint_coord()

Constructs a string containing a latitude or longitude in degrees and
minutes.
====================================================================== */

static char *sprint_coord( char *buf, double x )
{
   int m;
   
   m = ROUND( RADDEG( x ) * 60 );
   sprintf( buf, "%c%d\302\260%2.2d'",
      m < 0 ? '-' : '+', abs( m ) / 60, abs( m ) % 60 );
   return buf;
}


/*
======================================================================
tok_era()

Returns -1 if the era token begins with 'B' (eg "BC" or "BCE"), 1 if
it begins with 'A' or 'C' ("AD", "CE"), otherwise 0.
====================================================================== */

static int tok_era( char *tok )
{
   switch ( tok[ 0 ] ) {
      case 'B':
      case 'b':
         return -1;
      case 'A':
      case 'a':
      case 'C':
      case 'c':
         return 1;
      default:
         break;
   }
   return 0;
}


/*
======================================================================
tok_mon()

Matches the token with a month string.  Returns 1-12 for a matched
month name, otherwise 0.
====================================================================== */

static int tok_mon( char *tok )
{
   int i;
   char buf[ 4 ];
   
   buf[ 0 ] = toupper( tok[ 0 ] );
   buf[ 1 ] = tolower( tok[ 1 ] );
   buf[ 2 ] = tolower( tok[ 2 ] );
   buf[ 3 ] = 0;
   
   for ( i = 0; i < 12; i++ )
      if ( !strcmp( buf, month_str[ i ] ))
         return i + 1;
   return 0;
}


/*
======================================================================
parse_date()

Extract a JD from a date string typed by the user.

The canonical form of the date string is the output of sprint_date(),

   1 Jan 2000 AD

but a number of variations are supported.

   year
   month year
   year era
   day month year
   month day year
   month year era
   day month year era
   month day year era

The era may be BC or BCE, AD or CE.

If the year appears without the month, it may have a fractional part.
If it appears without the era, it may be negative.

Months must be given with at least their 3-letter abbreviations, not
numbers, and they must come before the year.
====================================================================== */

static int parse_date( char *buf, double *jd )
{
   static char *sep = " :;,/";
   char *tok[ 4 ];
   int i, d = 1, m = 1, era = 0;
   double y;

   for ( i = 0; i < 4; i++ ) {
      tok[ i ] = strtok( i ? NULL : buf, sep );
      if ( !tok[ i ] ) break;
   }

   switch ( i ) {
      case 1:
         y = atof( tok[ 0 ] );
         break;

      case 2:
         if ( isalpha( tok[ 0 ][ 0 ] )) {
            m = tok_mon( tok[ 0 ] );
            y = atoi( tok[ 1 ] );
         }
         else if ( isalpha( tok[ 1 ][ 0 ] )) {
            y = atof( tok[ 0 ] );
            era = tok_era( tok[ 1 ] );
         }
         break;

      case 3:
         if ( isalpha( tok[ 0 ][ 0 ] )) {
            m = tok_mon( tok[ 0 ] );
            if ( isalpha( tok[ 2 ][ 0 ] )) {
               y = atoi( tok[ 1 ] );
               era = tok_era( tok[ 2 ] );
            }
            else {
               d = atoi( tok[ 1 ] );
               y = atoi( tok[ 2 ] );
            }
         }
         else {
            d = atoi( tok[ 0 ] );
            m = tok_mon( tok[ 1 ] );
            y = atoi( tok[ 2 ] );
         }
         break;

      case 4:
         if ( isalpha( tok[ 0 ][ 0 ] )) {
            m = tok_mon( tok[ 0 ] );
            d = atoi( tok[ 1 ] );
         }
         else {
            d = atoi( tok[ 0 ] );
            m = tok_mon( tok[ 1 ] );
         }
         y = atoi( tok[ 2 ] );
         era = tok_era( tok[ 3 ] );
         break;

      default:
         break;
   }

   if ( d < 1 || d > 31 ||
        m < 1 || m > 12 ||
        (( int ) y ) == 0 )
      return 0;
        
   if ( era < 0 && y > 0 )
      y = -y;

   if ( y != ( int ) y )
      *jd = year_jd( y );
   else
      *jd = cal_jd( m, d, ( int ) y );
   return 1;
}


/*
======================================================================
parse_coord()

Extract a longitude or latitude from the string, written as either
degrees and minutes or fractional degrees.
====================================================================== */

static int parse_coord( char *buf, float *coord )
{
   static char *sep = " :;,/\302\260";
   char *tok[ 2 ];
   int neg;
   double d, m;
   
   tok[ 0 ] = strtok( buf, sep );
   if ( !tok[ 0 ] )
      return 0;
   tok[ 1 ] = strtok( NULL, sep );
   
   neg = tok[ 0 ][ 0 ] == '-';
   if ( tok[ 1 ] ) {
      d = fabs( atoi( tok[ 0 ] ));
      m = atoi( tok[ 1 ] ) / 60.0;
      if ( m >= 1.0 || m < 0.0 )
         return 0;
      d += m;
   }
   else
      d = fabs( atof( tok[ 0 ] ));

   d = ( int ) d + ROUND( 60 * ( d - ( int ) d )) / 60.0;
   d = DEGRAD( range( d, 360.0 ));
   if ( neg )
      d *= -1;

   *coord = ( float ) d;
   return 1;
}
   

/*
======================================================================
handle_epoch()

Reformat the string in the epoch entry when it loses focus.
====================================================================== */

static void handle_epoch( GtkWidget *entry, void *userdata )
{
   char buf[ 32 ];
   double jd;
   
   strcpy( buf, gtk_entry_get_text( GTK_ENTRY( entry )));
   if ( parse_date( buf, &jd ))
      gtk_entry_set_text( GTK_ENTRY( entry ), sprint_date( buf, jd ));
}

   
/*
======================================================================
handle_coord()

Reformat the string in one of the coordinate entries when it loses
focus.
====================================================================== */

static void handle_coord( GtkWidget *entry, void *userdata )
{
   char buf[ 16 ];
   float x;

   strcpy( buf, gtk_entry_get_text( GTK_ENTRY( entry )));
   if ( parse_coord( buf, &x ))
      gtk_entry_set_text( GTK_ENTRY( entry ), sprint_coord( buf, x ));
}
   
   
/*
======================================================================
handle_preset()

Set the text entries when one of the presets is chosen.
====================================================================== */

static void handle_preset( GtkWidget *combo, void *userdata )
{
   char buf[ 32 ];
   int i;
   
   i = gtk_combo_box_get_active( GTK_COMBO_BOX( combo_preset ));
   if ( i > 0 && i <= NPRESETS ) {
      --i;
      gtk_entry_set_text( GTK_ENTRY( entry_epoch ),
         sprint_date( buf, preset[ i ].jd ));
      gtk_entry_set_text( GTK_ENTRY( entry_lon ),
         sprint_coord( buf, preset[ i ].lon ));
      gtk_entry_set_text( GTK_ENTRY( entry_lat ),
         sprint_coord( buf, preset[ i ].lat ));
   }
}


/*
======================================================================
new_entry()

Create a text entry widget with a label.  The entry and label are
packed into an HBox, which in turn is packed into the container passed
as the third argument.
====================================================================== */

static GtkWidget *new_entry( char *name, int len, GtkWidget *container )
{
   GtkWidget *entry;
   GtkWidget *hbox;
   GtkWidget *label;
   
   hbox = gtk_hbox_new( FALSE, 4 );
   gtk_container_set_border_width( GTK_CONTAINER( hbox ), 8 );
   gtk_box_pack_start( GTK_BOX( container ), hbox, FALSE, FALSE, 0 );

   label = gtk_label_new( name );
   gtk_box_pack_start( GTK_BOX( hbox ), label, FALSE, FALSE, 0 );
   entry = gtk_entry_new_with_max_length( len );
   gtk_entry_set_width_chars( GTK_ENTRY( entry ), len / 2 );
   gtk_box_pack_start( GTK_BOX( hbox ), entry, FALSE, FALSE, 4 );
   
   return entry;
}


/*
======================================================================
is_preset()

Compares the epoch and lon/lat offsets to the elements of the preset
array.  If they match a preset, returns the array index, otherwise 0.
====================================================================== */

static int is_preset( double jd, float offsetLon, float offsetLat )
{
   int i;
   
   for ( i = 0; i < NPRESETS; i++ )
      if ( fabs( jd - preset[ i ].jd ) < 0.5 &&
           fabs( offsetLon - preset[ i ].lon ) < 0.000290888209 &&
           fabs( offsetLat - preset[ i ].lat ) < 0.000290888209 )
         return i + 1;
         
   return 0;
}


/*
======================================================================
new_preset()

Creates the preset combobox widget and label.  The combobox and label
are packed into an HBox, which is packed into the container.
====================================================================== */

static GtkWidget *new_preset( GtkWidget *container )
{
   GtkWidget *combo;
   GtkWidget *label;
   GtkWidget *hbox;
   int i;
   
   hbox = gtk_hbox_new( FALSE, 4 );
   gtk_container_set_border_width( GTK_CONTAINER( hbox ), 8 );
   gtk_box_pack_start( GTK_BOX( container ), hbox, FALSE, FALSE, 0 );

   label = gtk_label_new( "Presets" );
   gtk_box_pack_start( GTK_BOX( hbox ), label, FALSE, FALSE, 0 );

   combo = gtk_combo_box_new_text();
   gtk_combo_box_append_text( GTK_COMBO_BOX( combo ), "Custom" );
   for ( i = 0; i < NPRESETS; i++ )
      gtk_combo_box_append_text( GTK_COMBO_BOX( combo ), preset[ i ].name );
   gtk_box_pack_start( GTK_BOX( hbox ), combo, FALSE, FALSE, 4 );
   
   return combo;
}


/*
======================================================================
dlgEpoch()

Create and display a dialog that allows the user to set the epoch and
the longitude and latitude offsets for the Ptolemy catalogs.
====================================================================== */

int dlgEpoch( GtkWindow *window, double *jd, float *offsetLon, float *offsetLat )
{
   GtkWidget *dialog;
   GtkWidget *hbox;
   GtkWidget *widget;
   gint response;
   char buf[ 32 ];
   int i, ok = 0;
         
   /* the dialog */
   
   dialog = gtk_dialog_new_with_buttons(
      "Set Epoch",
      GTK_WINDOW( window ),
      GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
      GTK_STOCK_OK, GTK_RESPONSE_OK,
      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
      NULL );

   /* the epoch text field */
   
   entry_epoch = new_entry( "Epoch", 32, GTK_DIALOG( dialog )->vbox );

   /* the longitude and latitude offsets */
   
   widget = gtk_hseparator_new();
   gtk_box_pack_start( GTK_BOX( GTK_DIALOG( dialog )->vbox ), widget,
      TRUE, TRUE, 0 );
   widget = gtk_label_new( "Offsets added to catalog positions:" );
   gtk_box_pack_start( GTK_BOX( GTK_DIALOG( dialog )->vbox ), widget,
      FALSE, FALSE, 2 );
   hbox = gtk_hbox_new( TRUE, 4 );
   gtk_box_pack_start( GTK_BOX( GTK_DIALOG( dialog )->vbox ), hbox,
      FALSE, FALSE, 0 );

   entry_lon = new_entry( "Longitude", 16, hbox );
   entry_lat = new_entry( "Latitude", 16, hbox );

   /* the preset combobox */
   
   widget = gtk_hseparator_new();
   gtk_box_pack_start( GTK_BOX( GTK_DIALOG( dialog )->vbox ), widget,
      TRUE, TRUE, 0 );

   combo_preset = new_preset( GTK_DIALOG( dialog )->vbox );

   /* initialize the widget values */
   
   gtk_entry_set_text( GTK_ENTRY( entry_epoch ), sprint_date( buf, *jd ));
   gtk_entry_set_text( GTK_ENTRY( entry_lon ), sprint_coord( buf, *offsetLon ));
   gtk_entry_set_text( GTK_ENTRY( entry_lat ), sprint_coord( buf, *offsetLat ));
   gtk_combo_box_set_active( GTK_COMBO_BOX( combo_preset ),
      is_preset( *jd, *offsetLon, *offsetLat ));

   /* connect the event handlers */
   
   g_signal_connect( entry_epoch, "focus-out-event",
      G_CALLBACK( handle_epoch ), NULL );
   g_signal_connect( entry_lon,   "focus-out-event",
      G_CALLBACK( handle_coord ), NULL );
   g_signal_connect( entry_lat,   "focus-out-event",
      G_CALLBACK( handle_coord ), NULL );
   g_signal_connect( combo_preset, "changed",
      G_CALLBACK( handle_preset ), NULL );
   
   /* display the dialog */
   
   gtk_widget_show_all( GTK_DIALOG( dialog )->vbox );

   while ( 1 ) {
      response = gtk_dialog_run( GTK_DIALOG( dialog ));
      if ( response != GTK_RESPONSE_OK ) break;
   
      strcpy( buf, gtk_entry_get_text( GTK_ENTRY( entry_epoch )));
      if ( parse_date( buf, jd )) {
         strcpy( buf, gtk_entry_get_text( GTK_ENTRY( entry_lon )));
         if ( parse_coord( buf, offsetLon )) {
            strcpy( buf, gtk_entry_get_text( GTK_ENTRY( entry_lat )));
            if ( parse_coord( buf, offsetLat )) {
               ok = 1;
               break;
            }
         }
      }
      {
         GtkWidget *msgdlg = gtk_message_dialog_new( NULL, 0,
            GTK_MESSAGE_ERROR,
            GTK_BUTTONS_CLOSE,
            "One or more of the values is invalid." );
         gtk_dialog_run( GTK_DIALOG( msgdlg ));
         gtk_widget_destroy( msgdlg );
      }
   }

   gtk_widget_destroy( dialog );
   return ok;
}
