/*
======================================================================
files.c

Ernie Wright  10 Jan 07

File and color dialog support and file finder routine.

   fileRequest()
   findFile()
   selectColor()

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include "astro.h"


/*
======================================================================
fileRequest()

Prompt the user for a filename.
====================================================================== */

int fileRequest( GtkWindow *window, char *filename, char *buf, int bufsize,
   int type, char *caption )
{
   static char path[ MAX_PATH_LEN ];
   GtkWidget *dialog;
   int result;

   switch ( type ) {
      case FREQ_LOAD:
         dialog = gtk_file_chooser_dialog_new( caption, window,
            GTK_FILE_CHOOSER_ACTION_OPEN,
            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
            GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
            NULL );
         break;
      case FREQ_SAVE:
         dialog = gtk_file_chooser_dialog_new( caption, window,
            GTK_FILE_CHOOSER_ACTION_SAVE,
            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
            GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
            NULL);
         gtk_file_chooser_set_current_name( GTK_FILE_CHOOSER( dialog ),
            filename );
         break;
      default:
         return 0;
   }

   gtk_file_chooser_set_current_folder( GTK_FILE_CHOOSER( dialog ), path );

   result = gtk_dialog_run( GTK_DIALOG( dialog )) == GTK_RESPONSE_ACCEPT;
   
   if ( result ) {
      char *f, *p;
      f = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER( dialog ));
      p = gtk_file_chooser_get_current_folder( GTK_FILE_CHOOSER( dialog ));
      g_strlcpy( buf, f, bufsize );
      g_strlcpy( path, p, sizeof path );
      g_free( f );
      g_free( p );
   }

   gtk_widget_destroy( dialog );
   return result;
}


/*
======================================================================
append_node()

Concatenate a filename onto a path.
====================================================================== */

static void append_node( char *buf, char *node )
{
   int len;

   len = strlen( buf );
   if ( len && buf[ len - 1 ] != '/' ) {
      buf[ len ] = '/';
      buf[ len + 1 ] = 0;
   }
   strcat( buf, node );
}


/*
======================================================================
findFile()

Try to locate a file.  We look in

   1.  the current directory
   2.  the default data directory
   3.  the user's home directory

The default data directory, established by the configure script, is
typically /usr/local/share/almstars.  This is where we expect to find
the star catalog files and the help files.  PACKAGE_DATA_DIR is a
#define created on the compiler command line for this location.

The user's home directory is where we expect to find the config file.
Since this is the last place checked, it's the value stored in buf
when the function fails.  In the case of the config file, this is the
value that will be used to write the config when no config file was
found.
====================================================================== */

int findFile( GtkWindow *window, char *filename, char *buf, int bufsize )
{
   FILE *fp;

   strcpy( buf, filename );
   fp = fopen( buf, "r" );

   if ( !fp ) {
      char *p = PACKAGE_DATA_DIR;
      if ( strlen( p ) + strlen( filename ) + 2 < bufsize ) {
         strcpy( buf, PACKAGE_DATA_DIR );
         append_node( buf, filename );
         fp = fopen( buf, "r" );
      }
   }

   if ( !fp ) {
      char *p = getenv( "HOME" );
      if ( p && strlen( p ) + strlen( filename ) + 2 < bufsize ) {
         strcpy( buf, p );
         append_node( buf, filename );
         fp = fopen( buf, "r" );
      }
   }
         
   if ( fp ) {
      fclose( fp );
      return 1;
   }

   return 0;
}


/*
======================================================================
selectColor()

Opens a color dialog, initialized with the color argument.  If the
user chooses OK, the color argument receives the user's color choice.
Returns TRUE for OK, otherwise FALSE.
====================================================================== */

int selectColor( GtkWindow *window, COLORREF *color )
{
   GtkWidget *dialog;
   GdkColor gcolor;
   GtkColorSelection *cs;
   int result;
   
   dialog = gtk_color_selection_dialog_new( "Select Color" );
   cs = ( GtkColorSelection * )
      ((( GtkColorSelectionDialog * ) dialog )->colorsel );

   gcolor.red   = GetRValue16( *color );
   gcolor.green = GetGValue16( *color );
   gcolor.blue  = GetBValue16( *color );
   gtk_color_selection_set_current_color( cs, &gcolor );
      
   result = gtk_dialog_run( GTK_DIALOG( dialog )) == GTK_RESPONSE_OK;
   if ( result ) {
      gtk_color_selection_get_current_color( cs, &gcolor );
      *color = RGB16( gcolor.red, gcolor.green, gcolor.blue );
   }

   gtk_widget_destroy( dialog );
   return result;
}


