/*
======================================================================
info.c

Ernie Wright  29 Aug 06, 10 Jan 07

The Info dialog.  Shows a list of stars near a point in the diplay.

   clickInfo()

7 Jun 07    Fixed a -0 bug in the coordinate display.
            Added offsetLon and offsetLat to Ptolemy star coordinates.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include "astro.h"

#define TW 40   /* characters per row for the text */
#define TH 11   /* max number of rows */


extern GtkWidget *infoDialog;

static GtkWidget *label;
static char info[ 512 ];      /* at least TW x TH */
static int cx, cy;            /* point clicked by the user */
static int lw;                /* label width from make_label() */


/*
======================================================================
position_dlg()

Set the position of the dialog so that its corner is on the point
clicked by the user.  Ideally, we use the upper left corner, but we
use one of the others if, near as we can tell, part of the dialog
would be off-screen.
====================================================================== */

static void position_dlg( GtkWindow *window )
{
   GdkRectangle rw, rr;
   int x, y, ox, oy;

   gdk_window_get_frame_extents( GTK_WIDGET( window )->window, &rw );
   gdk_window_get_origin( GTK_WIDGET( window )->window, &ox, &oy );
   gtk_window_get_position( window, &x, &y );
   x += cx + ox - rw.x;
   y += cy + oy - rw.y;

   gdk_window_get_frame_extents( gdk_get_default_root_window(), &rr );
   if ( infoDialog->window ) {
      gdk_window_get_frame_extents( infoDialog->window, &rw );
      if ( x + rw.width  > rr.x + rr.width  ) x -= rw.width;
      if ( y + rw.height > rr.y + rr.height ) y -= rw.height;
   }
   else
      if ( x + lw > rr.x + rr.width ) x -= lw + 2 * ( ox - rw.x );
      
   gtk_window_move( GTK_WINDOW( infoDialog ), x, y );
}


/*
======================================================================
sprint_coords()

Write sexigesimal lat/lon to a string.
====================================================================== */

static void sprint_coords( char *buf, int lonm, int latm )
{
   sprintf( buf, "%3d\302\260%2.2d' %c%2d\xC2\xB0%2.2d'",
      lonm / 60,
      lonm % 60,
      latm < 0 ? '-' : ' ',
      abs( latm ) / 60,
      abs( latm ) % 60 );
}


/*
======================================================================
sprint_pcoords()

Call sprint_coords() with values in a PtolemyStar.  offsetLon and
offsetLat have to be treated like rotations.  You can't just add them
to the lon and lat and display the result, since it might be outside
the principal interval.
====================================================================== */

static void sprint_pcoords( char *buf, PtolemyStar *star, float offsetLon,
   float offsetLat )
{
   Vector pos;
   double lon, lat;
   int lonm, latm;
   
   lon = DEGRAD( star->lonZod * 30 + star->lonDeg + star->lonMin / 60.0 )
       + offsetLon;

   lat = DEGRAD( star->latDeg + star->latMin / 60.0 );
   if ( star->latNeg )
      lat = -lat;
   lat += offsetLat;

   sphere_vec( lon, lat, pos );
   vec_sphere( pos, &lon, &lat );
   
   lonm = ROUND( 60 * RADDEG( lon ));
   latm = ROUND( 60 * RADDEG( lat ));
   sprint_coords( buf, lonm, latm );
}


/*
======================================================================
sprint_ycoords()

Call sprint_coords() with values in a YaleStar.
====================================================================== */

static void sprint_ycoords( char *buf, YaleStar *star )
{
   double lon, lat;
   int lonm, latm;

   vec_sphere( star->pos, &lon, &lat );
   lonm = ROUND( RADDEG( lon ) * 60 );
   latm = ROUND( RADDEG( lat ) * 60 );
   sprint_coords( buf, lonm, latm );
}


/*
======================================================================
get_info()

Initialize the info[] string with information about the Ptolemy and
Yale stars near where the user clicked on the main window.
====================================================================== */

static int get_info( YaleStar star[], PtolemyConstellation *pcon,
   double pradius, float offsetLon, float offsetLat )
{
   char buf[ 128 ];
   int pf;
   int ny, yf[ 5 ];
   int i;


   /* pf is Ptolemy star index, or -1 if click is outside a Ptolemy circle
      ny is number of Yale stars */

   pf = clickPtolemy( pcon, cx, cy, pradius );
   ny = clickYale( star, cx, cy, pradius, yf );
   if ( pf < 0 && ny == 0 ) {
      sprintf( info, "(no stars)" );
      return 0;
   }

   info[ 0 ] = 0;

   if ( pf > -1 ) {

      /* Baily number, constellation name, index */

      sprintf( buf, "%d <b>%s %d</b>\r\n",
         pcon->star[ pf ].baily,
         pcon->name,
         pf + 1 );
      strcat( info, buf );

      /* Ptolemy's description */

      sprintf( buf, "%s\r\n", pcon->star[ pf ].description );
      strcat( info, buf );

      /* HR number of the Yale star identified with this Ptolemy entry */

      if ( pcon->star[ pf ].yaleIndex > 0 ) {
         sprintf( buf, "%4d", pcon->star[ pf ].yaleIndex );
         strcat( info, buf );
      }
      else
         strcat( info, "    " );

      /* Bayer or Flamsteed ID */

      sprintf( buf, " <b>%12s</b>  ", pcon->star[ pf ].id );
      strcat( info, buf );

      /* lon/lat coordinates */

      sprint_pcoords( buf, &pcon->star[ pf ], offsetLon, offsetLat );
      strcat( info, buf );

      /* magnitude, or "fnt." or "neb." */

      if ( pcon->star[ pf ].mag == 6.5f )
         strcat( info, "  fnt.\r\n" );
      else if ( pcon->star[ pf ].mag == 7.0f )
         strcat( info, "  neb.\r\n" );
      else {
         sprintf( buf, " %4.1f\r\n", pcon->star[ pf ].mag );
         strcat( info, buf );
      }
      strcat( info, "----------------------------------------\r\n" );
   }

   /* a line for each of the nearby Yale stars */

   for ( i = 0; i < ny; i++ ) {
      sprintf( buf, "%4d <b>%12s</b>  ", star[ yf[ i ]].hr, star[ yf[ i ]].name );
      strcat( info, buf );
      sprint_ycoords( buf, &star[ yf[ i ]] );
      strcat( info, buf );
      sprintf( buf, " %5.2f\r\n", star[ yf[ i ]].mag );
      strcat( info, buf );
   }

   return 1;
}


/*
======================================================================
make_label()

Create the label widget that will display the info string.
====================================================================== */

static void make_label( void )
{
   PangoFontDescription *fd;
   GtkRequisition r;
   char line[ TW + 1 ];
   
   label = gtk_label_new( NULL );
   gtk_container_add( GTK_CONTAINER( GTK_DIALOG( infoDialog )->vbox ), label );
   fd = pango_font_description_from_string( "Monospace 10" );
   gtk_widget_modify_font( label, fd );
   pango_font_description_free( fd );

/* this doesn't appear to do anything */
/* gtk_label_set_width_chars( GTK_LABEL( label ), TW ); */

   gtk_label_set_line_wrap( GTK_LABEL( label ), TRUE );
   gtk_label_set_selectable( GTK_LABEL( label ), TRUE );

   /* _really_ make the label TW characters wide */
   
   memset( line, '-', TW );
   line[ TW ] = 0;
   gtk_label_set_text( GTK_LABEL( label ), line );
   gtk_widget_show( label );
   gtk_widget_size_request( label, &r );
   gtk_widget_set_size_request( label, r.width, -1 );

   lw = r.width;
}


static void destroy( GtkWidget *widget, gpointer userdata )
{
   infoDialog = NULL;
}


/*
======================================================================
clickInfo()

Display a dialog containing information about stars near a user's
mouse click.

If x < 0, this is being called to update the contents of the dialog
without changing its position.  If the dialog isn't currently visible,
we can return without doing anything.
====================================================================== */

int clickInfo( GtkWindow *window, YaleStar star[], PtolemyConstellation *pcon,
   int x, int y, double pradius, float offsetLon, float offsetLat )
{
   
   if ( x < 0 && !infoDialog )
      return 0;
      
   /* the new dialog position */

   if ( x >= 0 ) {
      cx = x;
      cy = y;
   }

   /* if there are no nearby stars and the dialog isn't displayed,
      return without doing anything */
      
   if ( !get_info( star, pcon, pradius, offsetLon, offsetLat ) && !infoDialog )
      return 0;
      
   /* if the dialog doesn't exist, create it */
   
   if ( !infoDialog ) {
      infoDialog = gtk_dialog_new_with_buttons( "Nearest Stars", window,
         GTK_DIALOG_DESTROY_WITH_PARENT, "Close", GTK_RESPONSE_OK, NULL );
      make_label();
      g_signal_connect_swapped( infoDialog, "response",
         G_CALLBACK( gtk_widget_destroy ), infoDialog );
      g_signal_connect( infoDialog, "destroy", G_CALLBACK( destroy ), NULL );
   }

   gtk_label_set_markup( GTK_LABEL( label ), info );
   gtk_widget_show( label );
   position_dlg( window );
   gtk_window_present( GTK_WINDOW( infoDialog ));
   
   return 1;
}
