/*
======================================================================
menu.c

Ernie Wright  23 Jan 07

Create the popup menu and set menu item text and state.

   createMenu()
   menuSetLabel()
   menuSetCheck()

5 Jun 07    Added "Replace Current Catalog" and "Set Epoch" to the
            menu.
            
----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <gtk/gtk.h>
#include "menu.h"


/*
----------------------------------------------------------------------
The data for creating our popup menu is stored as an array of
MenuEntry structures.

The name is the text of the menu item.  The id is a Win32-ish control
ID used by the menu event handler to quickly identify the menu item.
The widget is the GtkMenuItem created for the item.  The handler is an
ID returned by g_signal_connect().  It's used in menuSetCheck() to
refer to the handler in g_signal_handler_block() and _unblock().

Some menu item ID ranges are set aside for menu items that should not
generate "activate" events, including separators and the parent items
of submenus.
---------------------------------------------------------------------- */

typedef struct {
   char *     name;
   int        id;
   GtkWidget *widget;
   gulong     handler;
} MenuEntry;

#define IDM_TOP  10
#define IDM_SEP  9000
#define IDM_MIN_ACTIVE  IDM_CONS
#define IDM_MAX_ACTIVE  IDM_SEP


static MenuEntry menuitem[] =
{
   /* Constellations:  North */
   { "Ursa Minor",         IDM_CONS +  1 },
   { "Ursa Major",         IDM_CONS +  2 },
   { "Draco",              IDM_CONS +  3 },
   { "Cepheus",            IDM_CONS +  4 },
   { "Bootes",             IDM_CONS +  5 },
   { "Corona Borealis",    IDM_CONS +  6 },
   { "Hercules",           IDM_CONS +  7 },
   { "Lyra",               IDM_CONS +  8 },
   { "Cygnus",             IDM_CONS +  9 },
   { "Cassiopeia",         IDM_CONS + 10 },
   { "Perseus",            IDM_CONS + 11 },
   { "Auriga",             IDM_CONS + 12 },
   { "Ophiuchus",          IDM_CONS + 13 },
   { "Serpens",            IDM_CONS + 14 },
   { "Sagitta",            IDM_CONS + 15 },
   { "Aquila",             IDM_CONS + 16 },
   { "Delphinus",          IDM_CONS + 17 },
   { "Equuleus",           IDM_CONS + 18 },
   { "Pegasus",            IDM_CONS + 19 },
   { "Andromeda",          IDM_CONS + 20 },
   { "Triangulum",         IDM_CONS + 21 },

   /* Constellations:  Zodiac */
   { "Aries",              IDM_CONS + 22 },
   { "Taurus",             IDM_CONS + 23 },
   { "Gemini",             IDM_CONS + 24 },
   { "Cancer",             IDM_CONS + 25 },
   { "Leo",                IDM_CONS + 26 },
   { "Virgo",              IDM_CONS + 27 },
   { "Libra",              IDM_CONS + 28 },
   { "Scorpius",           IDM_CONS + 29 },
   { "Sagittarius",        IDM_CONS + 30 },
   { "Capricornus",        IDM_CONS + 31 },
   { "Aquarius",           IDM_CONS + 32 },
   { "Pisces",             IDM_CONS + 33 },

   /* Constellations:  South */
   { "Cetus",              IDM_CONS + 34 },
   { "Orion",              IDM_CONS + 35 },
   { "Eridanus",           IDM_CONS + 36 },
   { "Lepus",              IDM_CONS + 37 },
   { "Canis Major",        IDM_CONS + 38 },
   { "Canis Minor",        IDM_CONS + 39 },
   { "Argo Navis",         IDM_CONS + 40 },
   { "Hydra",              IDM_CONS + 41 },
   { "Crater",             IDM_CONS + 42 },
   { "Corvus",             IDM_CONS + 43 },
   { "Centaurus",          IDM_CONS + 44 },
   { "Lupus",              IDM_CONS + 45 },
   { "Ara",                IDM_CONS + 46 },
   { "Corona Australis",   IDM_CONS + 47 },
   { "Piscis Austrinus",   IDM_CONS + 48 },

   /* Display Elements */
   { "1  Stars",                 IDM_LVIS + 1 },
   { "2  Circles",               IDM_LVIS + 2 },
   { "3  Numbers",               IDM_LVIS + 3 },
   { "4  Identification Lines",  IDM_LVIS + 4 },
   { "5  Ecliptic Grid",         IDM_LVIS + 5 },
   { "6  Equatorial Grid",       IDM_LVIS + 6 },

   /* Colors */
   { "Numbers",                  IDM_COLOR + 3 },
   { "Ecliptic Grid",            IDM_COLOR + 5 },
   { "Equatorial Grid",          IDM_COLOR + 6 },

   /* Arrow Keys */
   { "v  Change View Direction",    IDM_ARROWMODE + 1 },
   { "s  Shift Almagest Positions", IDM_ARROWMODE + 2 },
   { "---------------",             IDM_SEP + 1       },
   { "Default View",                IDM_ARROWMODE + 3 },
   { "Default Shift",               IDM_ARROWMODE + 4 },
   { "Zero Shift",                  IDM_ARROWMODE + 5 },

   /* Catalog */
   { "F1  Toomer/Grasshoff",       IDM_CAT + 1 },
   { "F2  Peters/Knobel",          IDM_CAT + 2 },
   { "F3  Manitius",               IDM_CAT + 3 },
   { "---------------",            IDM_SEP + 2 },
   { "c  Replace Current Catalog", IDM_CAT + 4 },
   { "e  Set Epoch",               IDM_CAT + 5 },

   /* top level items */
   { "Constellations:  North",   IDM_TOP + 1 },
   { "Constellations:  Zodiac",  IDM_TOP + 2 },
   { "Constellations:  South",   IDM_TOP + 3 },
   { "---------------",          IDM_SEP + 3 },
   { "Display Elements",         IDM_TOP + 4 },
   { "Colors",                   IDM_TOP + 5 },
   { "Arrow Keys",               IDM_TOP + 6 },
   { "Catalog",                  IDM_TOP + 7 },
   { "---------------",          IDM_SEP + 4 },
   { "User Guide",               IDM_HELP    },
   { "Print...",                 IDM_PRINT   },
   { "About",                    IDM_ABOUT   },
   { "Exit",                     IDM_QUIT    },

   { NULL },
};


/*
======================================================================
id_to_index()

Return the menuitem[] array index for the item with the given ID.
====================================================================== */

static int id_to_index( int id )
{
   int i;
   
   for ( i = 0; menuitem[ i ].name; i++ )
      if ( menuitem[ i ].id == id )
         return i;
   g_print( "couldn't find menu id %d\n", id );
   return -1;
}


/*
======================================================================
make_radios()

Create a group of GtkRadioMenuItems.  The group comprises the n
contiguous items in menuitem[] beginning with the one whose ID is
startid.  Internally to GTK, the items are linked into a list.  The
list is built by passing it in _item_new() and retrieving it in
_get_group() for each added item.
====================================================================== */

static void make_radios( int startid, int n )
{
   GSList *group = NULL;
   int start = id_to_index( startid );
   int i;
   
   for ( i = start; i < start + n; i++ ) {
      menuitem[ i ].widget = gtk_radio_menu_item_new_with_label(
         group, menuitem[ i ].name );
      group = gtk_radio_menu_item_get_group( GTK_RADIO_MENU_ITEM(
         menuitem[ i ].widget ));
   }
}


/*
======================================================================
make_checks()

Create n GtkCheckMenuItems beginning with the one whose ID is startid.
====================================================================== */

static void make_checks( int startid, int n )
{
   int start = id_to_index( startid );
   int i;
   
   for ( i = start; i < start + n; i++ )
      menuitem[ i ].widget = gtk_check_menu_item_new_with_label(
         menuitem[ i ].name );
}


/*
======================================================================
make_items()

Create n GtkMenuItems beginning with the one whose ID is startid.
====================================================================== */

static void make_items( int startid, int n )
{
   int start = id_to_index( startid );
   int i;

   for ( i = start; i < start + n; i++ )
      menuitem[ i ].widget = gtk_menu_item_new_with_label(
         menuitem[ i ].name );
}


/*
======================================================================
make_sep()

Create a GtkSeparatorMenuItem.
====================================================================== */

static void make_sep( int id )
{
   int index = id_to_index( id );
   menuitem[ index ].widget = gtk_separator_menu_item_new();
}


/*
======================================================================
make_menu()

Create a menu (which might be a submenu).  The menu will contain the n
menu items beginning with the one with ID startid.
====================================================================== */

static GtkWidget *make_menu( int startid, int n )
{
   GtkWidget *menu;
   int start = id_to_index( startid );
   int i;

   menu = gtk_menu_new();
   for ( i = start; i < start + n; i++ )
      gtk_menu_shell_append( GTK_MENU_SHELL( menu ), menuitem[ i ].widget );
   return menu;
}


/*
======================================================================
make_submenu()

Create a menu with make_menu() and attach it to the menu item
identified by id.
====================================================================== */

static GtkWidget *make_submenu( int id, int startid, int n )
{
   int index = id_to_index( id );
   GtkWidget *menu = make_menu( startid, n );
   gtk_menu_item_set_submenu( GTK_MENU_ITEM( menuitem[ index ].widget ), menu );
   return menu;
}


/*
======================================================================
createMenu()

Create the popup menu.

The menu_proc is attached to the "activate" signal for each menu item
(except separators and items with submenus).  The menu_proc callback
should have the prototype

   void menu_proc( GtkWidget *item, int id )

It receives the widget for the menu item and the unique integer ID
assigned to the item in the menuitem[] array.  It's called whenever
the user selects a menu item.  For check and radio items, it can also
be called whenever an item's state is changed programmatically.  When
a radio selection changes, for example, menu_proc is called for both
the old and new selections.  It will NOT be called for any item passed
to menuSetCheck(), which supresses the handler call.

The menu_open callback handles the event that occurs when the user
right-clicks the window.

   gint menu_open( GtkWidget *window, GdkEvent *event,
      GtkWidget *menu )

menu_open() should call gtk_menu_popup() to display the menu.  By
default, windows don't emit mouse clicks.  This needs to be set up by
calling

   gtk_widget_add_events( window, GDK_BUTTON_PRESS_MASK )

when the window is created.
====================================================================== */

GtkWidget *createMenu( GtkWidget *window, GCallback menu_proc,
   GCallback menu_open )
{
   GtkWidget *menu;
   int i;
   
   make_radios( IDM_CONS + 1, 48 );
   make_checks( IDM_LVIS + 1, 6 );   
   make_items(  IDM_COLOR + 3, 3 );
   make_radios( IDM_ARROWMODE + 1, 2 );
   make_sep(    IDM_SEP + 1 );
   make_items(  IDM_ARROWMODE + 3, 3 );
   make_radios( IDM_CAT + 1, 3 );
   make_sep(    IDM_SEP + 2 );
   make_items(  IDM_CAT + 4, 2 );
   make_items(  IDM_TOP + 1, 3 );
   make_sep(    IDM_SEP + 3 );
   make_items(  IDM_TOP + 4, 4 );
   make_sep(    IDM_SEP + 4 );
   make_items(  IDM_HELP, 4 );

   make_submenu( IDM_TOP + 1, IDM_CONS + 1, 21 );
   make_submenu( IDM_TOP + 2, IDM_CONS + 22, 12 );
   make_submenu( IDM_TOP + 3, IDM_CONS + 34, 15 );
   make_submenu( IDM_TOP + 4, IDM_LVIS + 1, 6 );
   make_submenu( IDM_TOP + 5, IDM_COLOR + 3, 3 );
   make_submenu( IDM_TOP + 6, IDM_ARROWMODE + 1, 6 );
   make_submenu( IDM_TOP + 7, IDM_CAT + 1, 6 );

   menu = make_menu( IDM_TOP + 1, 13 );
   
   for ( i = 0; menuitem[ i ].name; i++ )
      if ( menuitem[ i ].id > IDM_MIN_ACTIVE &&
           menuitem[ i ].id < IDM_MAX_ACTIVE )
         menuitem[ i ].handler = g_signal_connect( menuitem[ i ].widget,
            "activate", menu_proc, ( gpointer ) menuitem[ i ].id );

   gtk_widget_show_all( menu );
   g_signal_connect( window, "button-press-event", menu_open, menu );
   return menu;
}


/*
======================================================================
menuSetLabel()

Change the text of a menu item.
====================================================================== */

void menuSetLabel( int id, char *label )
{
   int index = id_to_index( id );
   if ( index < 0 ) return;
   gtk_label_set_text( GTK_LABEL( GTK_BIN( menuitem[ index ].widget )->child ),
      label );
}


/*
======================================================================
menuSetCheck()

Set the state of check and radio menu items.  Normally this would
cause the "activate" handler to be called, but we suppress that to
prevent a lot of spurious handler activity.
====================================================================== */

void menuSetCheck( int id, int state )
{
   int index = id_to_index( id );
   if ( index < 0 ) return;
   g_signal_handler_block(
      menuitem[ index ].widget,
      menuitem[ index ].handler );
   gtk_check_menu_item_set_active(
      GTK_CHECK_MENU_ITEM( menuitem[ index ].widget ), state );
   g_signal_handler_unblock(
      menuitem[ index ].widget,
      menuitem[ index ].handler );
}
