/*
======================================================================
precess.c

Ernie Wright  8 Jul 06

Rigorous precession calculation, along with related nutation and
obliquity calculations and conversion between ecliptic and equatorial
coordinates.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "astro.h"


/*
======================================================================
nutation()

Calculate nutation angles.

   t     number of Julian centuries since J2000.0
   dpsi  change in longitude (returned)
   deps  change in obliquity (returned)

Meeus (1998) p. 146.
====================================================================== */

void nutation( double t, double *dpsi, double *deps )
{
   static double t0 = -1e10, psi = 0.0, eps = 0.0;
   double lsun, lmoon, omega;

   if ( fabs( t - t0 ) < 1e-10 ) {
      *dpsi = psi;
      *deps = eps;
      return;
   }

   t0 = t;

   lsun  = DEGRAD( 280.4665  + 36000.7698  * t );
   lmoon = DEGRAD( 218.3165  + 481267.8813 * t );
   omega = DEGRAD( 125.04452 - 1934.136261 * t );

   psi = ASECRAD( -17.20 * sin( omega )
                  - 1.32 * sin( 2 * lsun )
                  - 0.23 * sin( 2 * lmoon )
                  + 0.21 * sin( 2 * omega ));

   eps = ASECRAD(   9.20 * cos( omega )
                  + 0.57 * cos( 2 * lsun )
                  + 0.10 * cos( 2 * lmoon )
                  - 0.09 * cos( 2 * omega ));

   *dpsi = psi;
   *deps = eps;
}


/*
======================================================================
obliq()

Return the true obliquity of the ecliptic at the given date.

Mean obliquity from Meeus (1998) p. 147.

Obliquity is the angle between the plane of the Earth's equator and
the plane of its orbit.  This calculation includes the nutation in
obliquity.
====================================================================== */

double obliq( double jd )
{
   static double cof[] = {
          2.45,
          5.79,
         27.87,
          7.12,
        -39.05,
       -249.67,
        -51.38,
       1999.25,
         -1.55,
      -4680.93
   };

   double ob, t, psi, eps;
   int j;

   t = ( jd - JD_J2000 ) / 3652500.0;

   /* get the nutation in obliquity (psi) */

   nutation( t * 100.0, &psi, &eps );

   /* calculate mean obliquity */

   ob = cof[ 0 ];
   for ( j = 1; j < 10; j++ )
      ob = ob * t + cof[ j ];

   /* return true obliquity */

   return DEGRAD( SEXDEC( 23, 26, 21.448 )) + psi + ASECRAD( ob ) * t;
}


/*
======================================================================
eq_ecl()

Multiply matrix M by a rotation that converts from equatorial to
ecliptic coordinates.  Duffett-Smith (1990) made clear to me that this
was simply a rotation around X by the obliquity of date.
====================================================================== */

void eq_ecl( double jd, Matrix3 M )
{
   mulrot3( M, obliq( jd ), PITCH );
}


/*
======================================================================
ecl_eq()

Multiply matrix M by a rotation that converts from ecliptic to
equatorial coordinates.
====================================================================== */

void ecl_eq( double jd, Matrix3 M )
{
   mulrot3( M, -obliq( jd ), PITCH );
}


/*
======================================================================
precess_ecl()

Precess ecliptic coordinates from J2000.0 to the input date.  Adapted
from Steve Moshier, Euler.txt and opreces.c, http://www.moshier.net/.
I've added nutation in longitude.

M is composed with (multiplied by) the rotations for precession, so it
must be initialized.  To precess in the opposite direction, from the
date to J2000.0, make M an identity matrix, call this function, then
transpose M before composing it with other transformations.

According to Moshier, the coefficients in pAc[] are from J. Laskar,
"Secular terms of classical planetary theories using the results of
general theory," Astronomy and Astrophysics 157, 59070 (1986).  The
coefficients in nodec[] and inclc[] are computed from Laskar's data as
explained in P. Bretagnon and G. Francou, "Planetary theories in
rectangular and spherical variables. VSOP87 solutions," Astronomy and
Astrophysics 202, 309-315 (1988).
====================================================================== */

void precess_ecl( double jd, Matrix3 M )
{
   static double pAc[] = {      /* coefficients for accumulated precession in */
      -8.66e-10,                /* longitude (arcsec) */
      -4.759e-8,
       2.424e-7,
       1.3095e-5,
       1.7451e-4,
      -1.8055e-3,
      -2.35316e-1,
       7.732e-2,
       1.111971e2,
       5.0290966e4
   };

   static double nodec[] = {    /* longitude of the ascending node of the */
       6.6402e-16,              /* ecliptic of date (radians) */
      -2.69151e-15,
      -1.547021e-12,
       7.521313e-12,
       6.3190131e-10,
      -3.48388152e-9,
      -1.813065896e-7,
       2.75036225e-8,
       7.4394531426e-5,
      -4.2078604317e-2,
       3.052112654975
   };

   static double inclc[] = {    /* inclination of the ecliptic (radians) */
       1.2147e-16,
       7.3759e-17,
      -8.26287e-14,
       2.503410e-13,
       2.4650839e-11,
      -5.4000441e-11,
       1.32115526e-9,
      -5.998737027e-7,
      -1.6242797091e-5,
       2.278495537e-3,
       0.0
   };

   double t, pA, i, omega, psi, eps;
   int j;

   /* thousands of years from J2000 */

   t = ( jd - JD_J2000 ) / 365250.0;

   /* accumulated precession in longitude */

   pA = pAc[ 0 ];
   for ( j = 1; j < 10; j++ )
      pA = pA * t + pAc[ j ];
   pA = ASECRAD( pA ) * t;

   /* ascending node of the moving ecliptic on the J2000 ecliptic */

   omega = nodec[ 0 ];
   for ( j = 1; j < 11; j++ )
      omega = omega * t + nodec[ j ];

   /* inclination of the moving ecliptic to the J2000 ecliptic */

   i = inclc[ 0 ];
   for ( j = 1; j < 11; j++ )
      i = i * t + inclc[ j ];

   /* get the nutation in longitude (eps) */

   nutation( t * 10.0, &psi, &eps );

   /* the transformation is a series of 3 rotations */

   mulrot3( M, omega, HEADING );
   mulrot3( M, i, PITCH );
   mulrot3( M, eps - ( omega + pA ), HEADING );
}


/*
======================================================================
precess2_ecl()

Precess ecliptic coordinates from and to arbitrary dates.
====================================================================== */

void precess2_ecl( double from, double to, Matrix3 M )
{
   Matrix3 P;

   precess_ecl( from, M );

   identity3( P );
   precess_ecl( to, P );
   transpose3( P );

   matmul3( M, P, M );
}
