/*
======================================================================
ptolemy.h

Ernie Wright  31 Aug 06

Definition of structures for a Ptolemy star and constellation, and
prototypes for reading and working on Almagest star catalogs.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include "vecmat.h"

#ifndef PTOLEMY_H
#define PTOLEMY_H

/*
----------------------------------------------------------------------
A PtolemyStar contains the information in a star entry in Ptolemy's
catalog.

Longitudes are given in a range from 0 to 30 degrees relative to the
start (smallest longitude) of a particular Zodiac constellation.

The 0 point of longitude is the vernal equinox, the point at which the
Sun (the ecliptic) crosses the celestial equator in northern Spring.
This is also the point at which Aries begins.

The epoch is uncertain.  Ptolemy compiled his catalog around 150 AD,
but Toomer notes that most of the longitudes are off by about a
degree, which would happen if Ptolemy applied his precession estimate
(1 degree per 100 years) to some or all of the star positions of
Hipparchus, who worked about 270 years earlier.

Longitudes and latitudes are given to a precision of one third, one
fourth, or one sixth of a degree, stored here as 20, 15, and 10
arcminutes.  Magnitudes are generally given as integers, but some
include the qualifiers "greater than" or "less than".  I've listed
these as plus or minus 0.3 magnitudes.
---------------------------------------------------------------------- */

typedef struct st_PtolemyStar {
   int    index;            /* index within constellation */
   int    lonZod;           /* zodiac constellation (longitude zone) */
   int    lonDeg;           /* ecliptic longitude (degrees) */
   int    lonMin;           /* longitude minutes */
   int    latNeg;           /* south latitude if true */
   int    latDeg;           /* ecliptic latitude (degrees) */
   int    latMin;           /* latitude minutes */
   float  mag;              /* magnitude */
   char   id[ 16 ];         /* modern identification */
   char  *description;      /* Ptolemy's identification by figure position */
   int    yaleIndex;        /* Yale Bright Star Catalog number */
   int    baily;            /* Baily number */
   float  lon;              /* longitude, radians */
   float  lat;              /* latitude, radians */
   Vector pos;              /* cartesian position vector */
   float  x;                /* image x coord */
   float  y;                /* image y coord */
   int    visible;          /* whether visible in the current view */
} PtolemyStar;


/* Ptolemy's star list is organized by constellation. */

typedef struct st_PtolemyConstellation {
   char        *name;       /* constellation name */
   PtolemyStar *star;       /* array of stars */
   int          inside;     /* number of stars in the figure */
   int          outside;    /* number of stars outside the figure */
   int          nstars;     /* inside + outside, for convenience */
} PtolemyConstellation;


/* a catalog is a named array of constellations */

#define NPCONS 48

typedef struct st_PtolemyCatalog {
   char                 *name;            /* descriptive name */
   char                 *filename;        /* data file */
   PtolemyConstellation con[ NPCONS ];    /* array of constellations */
} PtolemyCatalog;


PtolemyCatalog *readPCat( char *filename );
PtolemyCatalog *createPCat( char *filename, char *catname );
PtolemyCatalog *freePCat( PtolemyCatalog *pcat );
void freePCon( PtolemyConstellation *pcon );
void freePCons( PtolemyConstellation pcon[] );
void posPtolemyCon( PtolemyConstellation *pcon, float offsetLon,
        float offsetLat );
int  clickPtolemy( PtolemyConstellation *pcon, int x, int y, double pradius );
void getConViewDefaults( ConViewDefaults *cv, int i );
void initConViewDefaults( ConViewDefaults cv[] );


#endif
