/*
======================================================================
vecmat.c

Ernie Wright  07 Jan 04

Vector and matrix operations.
----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "astro.h"


#define EPSILON_F  1e-6
#define EPSILON_D  1e-15

/* some matrix shorthand */

#define m00  m[ 0 ][ 0 ]
#define m01  m[ 0 ][ 1 ]
#define m02  m[ 0 ][ 2 ]
#define m03  m[ 0 ][ 3 ]
#define m10  m[ 1 ][ 0 ]
#define m11  m[ 1 ][ 1 ]
#define m12  m[ 1 ][ 2 ]
#define m13  m[ 1 ][ 3 ]
#define m20  m[ 2 ][ 0 ]
#define m21  m[ 2 ][ 1 ]
#define m22  m[ 2 ][ 2 ]
#define m23  m[ 2 ][ 3 ]
#define m30  m[ 3 ][ 0 ]
#define m31  m[ 3 ][ 1 ]
#define m32  m[ 3 ][ 2 ]
#define m33  m[ 3 ][ 3 ]


void initm3( Matrix3 m,
   double a1, double b1, double c1,
   double a2, double b2, double c2,
   double a3, double b3, double c3 )
{
   m00 = a1;  m01 = b1;  m02 = c1;
   m10 = a2;  m11 = b2;  m12 = c2;
   m20 = a3;  m21 = b3;  m22 = c3;
}


void identity3( Matrix3 m )
{
   initm3( m,
      1.0, 0.0, 0.0,
      0.0, 1.0, 0.0,
      0.0, 0.0, 1.0 );
}


void transpose3( Matrix3 m )
{
   double temp;
   int i, j;

   for ( i = 0; i < 3; i++ )
      for ( j = 0; j < 3; j++ )
         if ( i != j ) {
            temp = m[ i ][ j ];
            m[ i ][ j ] = m[ j ][ i ];
            m[ j ][ i ] = temp;
         }
}


void copyv( Vector to, Vector from )
{
   memcpy( &to[ 0 ], &from[ 0 ], sizeof( Vector ));
}


void copym3( Matrix3 to, Matrix3 from )
{
   memcpy( &to[ 0 ][ 0 ], &from[ 0 ][ 0 ], sizeof( Matrix3 ));
}


double dot( Vector a, Vector b )
{
   return a[ 0 ] * b[ 0 ] + a[ 1 ] * b[ 1 ] + a[ 2 ] * b[ 2 ];
}


double vlenSquared( Vector a )
{
   return dot( a, a );
}


double vlen( Vector a )
{
   return sqrt( vlenSquared( a ));
}


void scalev( Vector a, double s )
{
   a[ 0 ] *= s;
   a[ 1 ] *= s;
   a[ 2 ] *= s;
}


void setvlen( Vector a, double d )
{
   double d0 = vlen( a );

   if ( d0 != 0.0 ) scalev( a, d / d0 );
}


void normalize( Vector a )
{
   setvlen( a, 1.0 );
}


/*
======================================================================
ry_heading()

Convert between LW heading and astronomical angles around the up axis,
(e.g. right ascension, ecliptic longitude).  Works both ways.

LW heading = 0 corresponds to +Z and increases clockwise (viewed from
above).  R.A. and longitude = 0 correspond to +X and increase CCW.
====================================================================== */

double ry_heading( double a )
{
   double b;

   b = HALFPI - a;
   if ( b < 0 ) b += TWOPI;
   return b;
}


/*
======================================================================
rx_pitch()

Convert between LW pitch and astronomical angles (e.g. declination,
ecliptic latitude) around the X axis (the axis pointing towards 0 R.A.
or ecliptic longitude).  Works both ways.

We just need to account for positive angles being in opposite
directions.
====================================================================== */

double rx_pitch( double a )
{
   return -a;
}


/*
======================================================================
range()

Constrain a periodic value v to the range [0, r).
====================================================================== */

double range( double v, double r )
{
   return v - r * floor( v / r );
}


/*
======================================================================
sphere_vec()

Convert astro spherical coordinates into a LW unit direction vector.
====================================================================== */

void sphere_vec( double lon, double lat, Vector a )
{
   double h, p, ch, sh, cp, sp;

   h = ry_heading( lon );
   p = rx_pitch( lat );

   ch = cos( h );
   sh = sin( h );
   cp = cos( p );
   sp = sin( p );

   a[ 0 ] = sh * cp;
   a[ 1 ] = -sp;
   a[ 2 ] = ch * cp;
}


/*
======================================================================
vec_sphere()

Convert a LW direction vector into astro spherical coordinates.
====================================================================== */

void vec_sphere( Vector a, double *lon, double *lat )
{
   Vector n;
   double h, p;

   copyv( n, a );
   normalize( n );

   p = asin( -n[ 1 ] );

   if ( n[ 0 ] == 0.0 && n[ 2 ] == 0.0 )
      h = 0.0;
   else {
      h = atan2( n[ 0 ], n[ 2 ] );
      if ( h < 0.0 ) h += TWOPI;
   }

   *lon = ry_heading( h );
   *lat = rx_pitch( p );
}


/*
======================================================================
transform()

Multiply a vector by a matrix, storing the result in a second vector
(which may actually be the same as the first).
====================================================================== */

void transform( Vector a, Matrix3 m, Vector b )
{
   Vector c;
   int i;

   for ( i = 0; i < 3; i++ )
      c[ i ] = a[ 0 ] * m[ 0 ][ i ]
             + a[ 1 ] * m[ 1 ][ i ]
             + a[ 2 ] * m[ 2 ][ i ];
   copyv( b, c );
}


/*
======================================================================
matmul3()

Multiply two matrices, storing the result in a third (which may
actually be one of the other two).
====================================================================== */

void matmul3( Matrix3 a, Matrix3 b, Matrix3 c )
{
   int i, j, k;
   Matrix3 m = { 0.0 };

   for ( i = 0; i < 3; i++ )
      for ( j = 0; j < 3; j++ )
         for ( k = 0; k < 3; k++ )
            m[ i ][ j ] += a[ i ][ k ] * b[ k ][ j ];
   copym3( c, m );
}


/*
======================================================================
mulrot3()

Multiply a matrix by a rotation around an axis.
====================================================================== */

void mulrot3( Matrix3 M, double angle, int axis )
{
   Matrix3 R;
   double s, c;

   s = sin( angle );
   c = cos( angle );

   switch ( axis ) {
      case HEADING:
         initm3( R,  c , 0.0, -s ,
                    0.0, 1.0, 0.0,
                     s , 0.0,  c );
         break;
      case PITCH:
         initm3( R, 1.0, 0.0, 0.0,
                    0.0,  c ,  s ,
                    0.0, -s ,  c );
         break;
      case BANK:
         initm3( R,  c ,  s , 0.0,
                    -s ,  c , 0.0,
                    0.0, 0.0, 1.0 );
         break;
      default:
         return;
   }

   matmul3( M, R, M );
}


void transform4( Vector a, Matrix4 m, Vector b )
{
   double w;
   int i;

   for ( i = 0; i < 3; i++ )
      b[ i ] = a[ 0 ] * m[ 0 ][ i ]
             + a[ 1 ] * m[ 1 ][ i ]
             + a[ 2 ] * m[ 2 ][ i ] + m[ 3 ][ i ];

   w = a[ 0 ] * m[ 0 ][ 3 ]
     + a[ 1 ] * m[ 1 ][ 3 ]
     + a[ 2 ] * m[ 2 ][ 3 ] + m[ 3 ][ 3 ];

   if ( w != 0.0 ) {
      b[ 0 ] /= w;
      b[ 1 ] /= w;
      b[ 2 ] /= w;
   }
}


void matmul4( Matrix4 a, Matrix4 b, Matrix4 c )
{
   int i, j, k;
   Matrix4 m = { 0.0 };

   for ( i = 0; i < 4; i++ )
      for ( j = 0; j < 4; j++ )
         for ( k = 0; k < 4; k++ )
            m[ i ][ j ] += a[ i ][ k ] * b[ k ][ j ];

   memcpy( &c[ 0 ][ 0 ], &m[ 0 ][ 0 ], sizeof( Matrix4 ));
}


void scalem3( Matrix3 a, double s )
{
   int i, j;

   for ( i = 0; i < 3; i++ )
      for ( j = 0; j < 3; j++ )
         a[ i ][ j ] *= s;
}


void scalem4( Matrix4 a, double s )
{
   int i, j;

   for ( i = 0; i < 4; i++ )
      for ( j = 0; j < 4; j++ )
         a[ i ][ j ] *= s;
}


double det2( double a, double b, double c, double d )
{
    return a * d - b * c;
}


double det3( Matrix3 m )
{
   return   m00 * det2( m11, m21, m12, m22 )
          - m01 * det2( m10, m20, m12, m22 )
          + m02 * det2( m10, m20, m11, m21 );
}


double det4( Matrix4 m )
{
   Matrix3 p;
   double a, b, c, d;

   initm3( p, m11, m21, m31, m12, m22, m32, m13, m23, m33 );  a = det3( p );
   initm3( p, m10, m20, m30, m12, m22, m32, m13, m23, m33 );  b = det3( p );
   initm3( p, m10, m20, m30, m11, m21, m31, m13, m23, m33 );  c = det3( p );
   initm3( p, m10, m20, m30, m11, m21, m31, m12, m22, m32 );  d = det3( p );

   return m00 * a - m01 * b + m02 * c - m03 * d;
}


void adjoint3( Matrix3 m, Matrix3 adj )
{
   adj[ 0 ][ 0 ] =  det2( m11, m12, m21, m22 );
   adj[ 1 ][ 0 ] = -det2( m10, m12, m20, m22 );
   adj[ 2 ][ 0 ] =  det2( m10, m11, m20, m21 );
   adj[ 0 ][ 1 ] = -det2( m01, m02, m21, m22 );
   adj[ 1 ][ 1 ] =  det2( m00, m02, m20, m22 );
   adj[ 2 ][ 1 ] = -det2( m00, m01, m20, m21 );
   adj[ 0 ][ 2 ] =  det2( m01, m02, m11, m12 );
   adj[ 1 ][ 2 ] = -det2( m00, m02, m10, m12 );
   adj[ 2 ][ 2 ] =  det2( m00, m01, m10, m11 );
}


void adjoint4( Matrix4 m, Matrix4 adj )
{
   Matrix3 a;

   initm3( a, m11, m21, m31, m12, m22, m32, m13, m23, m33 );  adj[ 0 ][ 0 ] =  det3( a );
   initm3( a, m10, m20, m30, m12, m22, m32, m13, m23, m33 );  adj[ 1 ][ 0 ] = -det3( a );
   initm3( a, m10, m20, m30, m11, m21, m31, m13, m23, m33 );  adj[ 2 ][ 0 ] =  det3( a );
   initm3( a, m10, m20, m30, m11, m21, m31, m12, m22, m32 );  adj[ 3 ][ 0 ] = -det3( a );
   initm3( a, m01, m21, m31, m02, m22, m32, m03, m23, m33 );  adj[ 0 ][ 1 ] = -det3( a );
   initm3( a, m00, m20, m30, m02, m22, m32, m03, m23, m33 );  adj[ 1 ][ 1 ] =  det3( a );
   initm3( a, m00, m20, m30, m01, m21, m31, m03, m23, m33 );  adj[ 2 ][ 1 ] = -det3( a );
   initm3( a, m00, m20, m30, m01, m21, m31, m02, m22, m32 );  adj[ 3 ][ 1 ] =  det3( a );
   initm3( a, m01, m11, m31, m02, m12, m32, m03, m13, m33 );  adj[ 0 ][ 2 ] =  det3( a );
   initm3( a, m00, m10, m30, m02, m12, m32, m03, m13, m33 );  adj[ 1 ][ 2 ] = -det3( a );
   initm3( a, m00, m10, m30, m01, m11, m31, m03, m13, m33 );  adj[ 2 ][ 2 ] =  det3( a );
   initm3( a, m00, m10, m30, m01, m11, m31, m02, m12, m32 );  adj[ 3 ][ 2 ] = -det3( a );
   initm3( a, m01, m11, m21, m02, m12, m22, m03, m13, m23 );  adj[ 0 ][ 3 ] = -det3( a );
   initm3( a, m00, m10, m20, m02, m12, m22, m03, m13, m23 );  adj[ 1 ][ 3 ] =  det3( a );
   initm3( a, m00, m10, m20, m01, m11, m21, m03, m13, m23 );  adj[ 2 ][ 3 ] = -det3( a );
   initm3( a, m00, m10, m20, m01, m11, m21, m02, m12, m22 );  adj[ 3 ][ 3 ] =  det3( a );
}


void inverse3( Matrix3 m, Matrix3 inv )
{
   double det;

   det = det3( m );
   if ( fabs( det ) < EPSILON_D ) return;

   adjoint3( m, inv );
   scalem3( inv, 1.0 / det );
}


void inverse4( Matrix4 m, Matrix4 inv )
{
   double det;

   det = det4( m );
   if ( fabs( det ) < EPSILON_D ) return;

   adjoint4( m, inv );
   scalem4( inv, 1.0 / det );
}
