/*
======================================================================
view.c

Ernie Wright  10 Jul 06

Functions related to the camera model used to create the display.

   viewInit()     set some View fields
   viewUpdate()   calculate pradius and transformation matrices
   viewToPixel()  convert from camera to pixel (x, y)
   pixelToEcl()   convert from pixel (x, y) to lon-lat

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <math.h>
#include "astro.h"


/*
======================================================================
make_viewmat()

Create the 4x4 viewing transformation matrix and its inverse.

The camera's orientation is defined by ecliptic longitude/latitude and
a LW bank angle.  After the astronomical angles are converted to LW
heading and pitch, the three Euler angles determine the axes of a
left-handed LW camera, and these become the columns of the matrix.

The inverse matrix is used to get back to ecliptic lon/lat from screen
coordinates.  It's the transpose of the upper left 3x3 part of the
viewing matrix, with a fourth column of [0, 0, -1, zoom].

Actually inverting the matrix *almost* works, but it's numerically
unstable.  You often get tiny (on the order of 1e-15) non-zero values
that should really be 0.0, and these blow up transformed points or
flip their signs.  So the inverse is written by hand.
====================================================================== */

static void make_viewmat( View *view )
{
   double h, p;
   double sh, ch, sp, cp, sb, cb;

   h = ry_heading( view->lon );
   p = rx_pitch( view->lat );

   sh = sin( h );
   ch = cos( h );
   sp = sin( p );
   cp = cos( p );
   sb = sin( view->bank );
   cb = cos( view->bank );

   view->Mv[ 0 ][ 0 ] =  cb * ch + sb * sp * sh;
   view->Mv[ 0 ][ 1 ] = -sb * ch + cb * sp * sh;
   view->Mv[ 0 ][ 2 ] =  cp * sh;
   view->Mv[ 0 ][ 3 ] =  view->Mv[ 0 ][ 2 ] / view->zoom;
   view->Mv[ 1 ][ 0 ] =  sb * cp;
   view->Mv[ 1 ][ 1 ] =  cb * cp;
   view->Mv[ 1 ][ 2 ] = -sp;
   view->Mv[ 1 ][ 3 ] =  view->Mv[ 1 ][ 2 ] / view->zoom;
   view->Mv[ 2 ][ 0 ] = -sh * cb + sb * sp * ch;
   view->Mv[ 2 ][ 1 ] =  sb * sh + cb * sp * ch;
   view->Mv[ 2 ][ 2 ] =  cp * ch;
   view->Mv[ 2 ][ 3 ] =  view->Mv[ 2 ][ 2 ] / view->zoom;
   view->Mv[ 3 ][ 0 ] =  0.0;
   view->Mv[ 3 ][ 1 ] =  0.0;
   view->Mv[ 3 ][ 2 ] =  0.0;
   view->Mv[ 3 ][ 3 ] =  1.0 / view->zoom;

   view->iMv[ 0 ][ 0 ] = view->Mv[ 0 ][ 0 ];
   view->iMv[ 0 ][ 1 ] = view->Mv[ 1 ][ 0 ];
   view->iMv[ 0 ][ 2 ] = view->Mv[ 2 ][ 0 ];
   view->iMv[ 0 ][ 3 ] = 0.0;
   view->iMv[ 1 ][ 0 ] = view->Mv[ 0 ][ 1 ];
   view->iMv[ 1 ][ 1 ] = view->Mv[ 1 ][ 1 ];
   view->iMv[ 1 ][ 2 ] = view->Mv[ 2 ][ 1 ];
   view->iMv[ 1 ][ 3 ] = 0.0;
   view->iMv[ 2 ][ 0 ] = view->Mv[ 0 ][ 2 ];
   view->iMv[ 2 ][ 1 ] = view->Mv[ 1 ][ 2 ];
   view->iMv[ 2 ][ 2 ] = view->Mv[ 2 ][ 2 ];
   view->iMv[ 2 ][ 3 ] = -1.0;
   view->iMv[ 3 ][ 0 ] = 0.0;
   view->iMv[ 3 ][ 1 ] = 0.0;
   view->iMv[ 3 ][ 2 ] = 0.0;
   view->iMv[ 3 ][ 3 ] = view->zoom;
}


/*
======================================================================
viewInit()

Set the view parameters.
====================================================================== */

void viewInit( View *view, double lon, double lat, double bank, double zoom )
{
   view->lon = lon;
   view->lat = lat;
   view->bank = bank;
   view->zoom = zoom;
}


/*
======================================================================
viewUpdate()

Recalculate the pradius and viewing transformation after one of the
view parameters has changed.
====================================================================== */

void viewUpdate( View *view )
{
   view->pradius = view->height * DEGRAD( 1.0 )
      / ( 2.0 * atan2( 2.0, view->zoom ));
   make_viewmat( view );
}


/*
======================================================================
viewToPixel()

Convert from camera coordinates to pixel (x, y).  Returns TRUE if the
coordinates are visible (are inside the window), otherwise FALSE.
====================================================================== */

int viewToPixel( View *view, Vector pos, float *x, float *y )
{
   double s;

   if ( pos[ 2 ] <= 0.0 ) return 0;

   s = view->height * 0.5;
   *y = ( float )( s - pos[ 1 ] * s );
   *x = ( float )( view->width * 0.5 + pos[ 0 ] * s );

   return (( *x >= 0 ) && ( *x < view->width )
        && ( *y >= 0 ) && ( *y < view->height ));
}


/*
======================================================================
pixelToEcl()

Given pixel coordinates, return the longitude and latitude.  The pixel
(x, y) is converted to camera coordinates, then multiplied by the
inverse of the viewing transformation, and finally converted from
Cartesian to spherical astronomical coordinates.
====================================================================== */

void pixelToEcl( View *view, int x, int y, float *lon, float *lat )
{
   Vector pos, tpos;
   double dlon, dlat, s;

   s = view->height * 0.5;
   pos[ 0 ] = ( x - view->width * 0.5 ) / s;
   pos[ 1 ] = ( s - y ) / s;
   s = pos[ 0 ] * pos[ 0 ] + pos[ 1 ] * pos[ 1 ];
   pos[ 2 ] = 0.5 * ( view->zoom - s / view->zoom );

   transform4( pos, view->iMv, tpos );
   normalize( tpos );
   vec_sphere( tpos, &dlon, &dlat );
   *lon = ( float ) dlon;
   *lat = ( float ) dlat;
}


/*
======================================================================
pixelToEclEq()

Given pixel coordinates, return both the ecliptic and equatorial
coordinates.  The pixel (x, y) is converted to ecliptic coordinates,
which are then converted to equatorial.
====================================================================== */

void pixelToEclEq( View *view, int x, int y, double jd,
   float *lon, float *lat, float *ra, float *dec )
{
   Matrix3 M;
   double dra, ddec, pos[ 3 ];

   pixelToEcl( view, x, y, lon, lat );
   
   identity3( M );
   ecl_eq( jd, M );
   sphere_vec( *lon, *lat, pos );
   transform( pos, M, pos );
   vec_sphere( pos, &dra, &ddec );
   *ra  = ( float ) dra;
   *dec = ( float ) ddec;
}
