/*
======================================================================
window.c

Ernie Wright  08 Jul 06

Entry point and user interface handling for Almagest Stars.

5 Jun 07    Updates for handling catalog replacement and setting the
            epoch.  Fixed -0 bugs in the coordinate display.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>  /* for execlp() in show_help() */
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "astro.h"
#include "menu.h"

/* argument to change_view() */

#define PARAM_LON 1
#define PARAM_LAT 2
#define PARAM_BANK 3
#define PARAM_ZOOM 4

/* graphics context required by GDK functions */

#define GC_FG drawarea->style->fg_gc[ 0 ]

static YaleStar star[ 9096 ];
static PtolemyCatalog *pcat[ 3 ];
static PtolemyConstellation *pcon = NULL;
static double jd;
static char *catname;
static char cfgname[ MAX_PATH_LEN ];
static char *helpcmd[ 2 ];
static ConViewDefaults conview[ 48 ];
static View view;
static float errLon, errLat;
static float offsetLon, offsetLat;
static int arrowmode;
static int layervis = 31;
static int curcon = 0;
static int undo = 1;

static GtkWidget *window;
static GtkWidget *drawarea;
static GdkPixmap *pixmap;
static PangoLayout *layout;

/* global pointer to the Info dialog, used by key_snoop() */

GtkWidget *infoDialog;


/*
======================================================================
load_yale()

Load the bright star catalog.  findFile() finds the path to the
catalog.  posYale() reduces the star positions to the epoch set by the
user.  By default, this is the epoch of the Almagest--July 20, 137 AD.
====================================================================== */

static int load_yale( void )
{
   char buf[ MAX_PATH_LEN ];

   if ( !findFile( NULL, "catb.bin", buf, sizeof( buf ))) return 0;
   if ( !readYale( buf, star )) return 0;
   posYale( star, jd );
   return 1;
}


/*
======================================================================
load_ptolemy()

Load an Almagest catalog.  findFile() finds the path to the catalog.
====================================================================== */

static int load_ptolemy( char *filename, int i )
{
   char buf[ MAX_PATH_LEN ];

   if ( !findFile( NULL, filename, buf, sizeof( buf ))) return 0;
   if ( !( pcat[ i ] = readPCat( buf ))) return 0;
   return 1;
}


/*
======================================================================
get_conview()
put_conview()

Store and retrieve default viewing parameters.
====================================================================== */

static void get_conview( ConViewDefaults *cv )
{
   view.lon  = cv->lon;
   view.lat  = cv->lat;
   view.bank = cv->bank;
   view.zoom = cv->zoom;
   errLon    = cv->errLon;
   errLat    = cv->errLat;
}


static void put_conview( ConViewDefaults *cv )
{
   cv->lon    = ( float ) view.lon;
   cv->lat    = ( float ) view.lat;
   cv->bank   = ( float ) view.bank;
   cv->zoom   = ( float ) view.zoom;
   cv->errLon = errLon;
   cv->errLat = errLat;
}


/*
======================================================================
init_stars()

Initialize drawbuf.c, load the catalogs, and initialize the view.
This only needs to be called once, at the start of the program.
====================================================================== */

static int init_stars( void )
{
   initDraw();

   if ( !load_yale() ) return 0;
   if ( !load_ptolemy( "cat1.dat", 0 )) return 0;
   if ( !load_ptolemy( "cat2.dat", 1 )) return 0;
   if ( !load_ptolemy( "cat3.dat", 2 )) return 0;
   pcon = pcat[ 0 ]->con;
   catname = pcat[ 0 ]->name;

   get_conview( &conview[ curcon ] );
   viewUpdate( &view );

   return 1;
}


/*
======================================================================
coord_text()

Format and display coordinate text.  Used by draw_mouse().
====================================================================== */

static void coord_text( float u, float v, int ly, COLORREF c )
{
   char buf[ 64 ];
   int ud, um, vd, vm, len, x, y, w, h;

   um = abs(( int )( 60 * u ));
   vm = abs( ROUND( 60 * v ));
   ud = um / 60;
   vd = vm / 60;
   um = abs( um - 60 * ud );
   vm = abs( vm - 60 * vd );

   len = sprintf( buf, "    %d\302\260%2.02d' %c%d\302\260%2.02d'",
      ud, um, v < 0 ? '-' : ' ', vd, vm );

   setPangoColor( layout, c, RGB( 0, 0, 0 ));
   pango_layout_set_text( layout, buf, len );
   pango_layout_get_pixel_size( layout, &w, &h );

   x = view.width - ( w + h );
   y = view.height - ( 3 + 2 * ly ) * h / 2;
   gdk_draw_layout( pixmap, GC_FG, x, y, layout );
   gtk_widget_queue_draw_area( drawarea, x, y, w, h );
}


/*
======================================================================
draw_mouse()

Display the ecliptic longitude and latitude corresponding to the
current position of the mouse cursor.  If displaying the ra/dec grid,
also write the mouse ra/dec.
====================================================================== */

static void draw_mouse( int x, int y )
{
   float lon, lat, ra, dec;

   if ( layervis & LVIS_RADECLN ) {
      pixelToEclEq( &view, x, y, jd, &lon, &lat, &ra, &dec );
      ra  = ( float ) RADDEG( ra );
      dec = ( float ) RADDEG( dec );
   }
   else
      pixelToEcl( &view, x, y, &lon, &lat );

   lon = ( float ) RADDEG( lon );
   lat = ( float ) RADDEG( lat );

   coord_text( lon, lat, 0, RGB( 255, 255, 128 ));
   if ( layervis & LVIS_RADECLN )
      coord_text( ra, dec, 1, getPrefColor( PREFCOLOR_RADECTXT ));
}


/*
======================================================================
get_mousepos()

Find the window coordinates of the current position of the mouse
cursor.

When draw_mouse() is called in response to a mouse motion event, we
already know where the mouse is.  But the ecliptic coordinates of the
mouse position can change without the mouse moving, and in those
cases we need a way to find out where the mouse is before calling
draw_mouse().
====================================================================== */

static int get_mousepos( int *x, int *y )
{
   GdkModifierType state;
   
   if ( !gdk_window_get_pointer( window->window, x, y, &state ))
      return 0;
   return *x >= 0 && *x < view.width && *y >= 0 && *y < view.height;
}


/*
======================================================================
draw_status()

Draw a line of text on the display showing the current constellation,
view parameters, systematic error (shift) amounts, arrow key mode, and
Almagest catalog version.
====================================================================== */

static void draw_status( void )
{
   char buf[ 160 ];
   int x, y, w, h, len;

   len = setStatusText( buf, &view, pcon[ curcon ].name, errLon, errLat,
      catname, arrowmode );
   setPangoColor( layout, RGB( 255, 255, 128 ), 1 );
   pango_layout_set_text( layout, buf, len );
   pango_layout_get_pixel_size( layout, &w, &h );

   x = 10;
   y = view.height - 3 * h / 2;
   gdk_draw_layout( pixmap, GC_FG, x, y, layout );
   gtk_widget_queue_draw_area( drawarea, x, y, w, h );

   if ( get_mousepos( &x, &y ))
      draw_mouse( x, y );
}


/*
======================================================================
init_drawinfo()

Write our current state in a DrawInfo so that we can draw or save the
display.
====================================================================== */

static void init_drawinfo( DrawInfo *di, char *filename )
{
   di->star      = star;
   di->pcon      = &pcon[ curcon ];
   di->errLon    = errLon + offsetLon;
   di->errLat    = errLat + offsetLat;
   di->view      = &view;
   di->layervis  = layervis;
   di->catname   = catname;
   di->filename  = filename;
   di->jd        = jd;
}


/*
======================================================================
draw_stars()

Update the display to reflect our current state.
====================================================================== */

static void draw_stars( void )
{
   DrawInfo di;

   init_drawinfo( &di, NULL );
   drawDBuf( &di, pixmap, GC_FG, layout );
   draw_status();
   gtk_widget_queue_draw( drawarea );
}


/*
======================================================================
pdf_stars()

Save the current view as a PDF.
====================================================================== */

static void pdf_stars( void )
{
   DrawInfo di;
   char filename[ MAX_PATH_LEN ];
   char suggest[ 40 ];
   int i, len;
   
   len = sprintf( suggest, "%s.pdf", pcon[ curcon ].name );
   for ( i = 0; i < len; i++ )
      if ( suggest[ i ] == ' ' )
         suggest[ i ] = '_';
         
   if ( !fileRequest( GTK_WINDOW( window ), suggest, filename,
      sizeof( filename ), FREQ_SAVE, "Save View As PDF" ))
      return;

   init_drawinfo( &di, filename );
   drawPDF( &di );
}


/*
======================================================================
change_arrowmode()

Change the meaning of an arrow key press.
====================================================================== */

static void change_arrowmode( int newmode )
{
   arrowmode = newmode;
   draw_status();
}


/*
======================================================================
change_view()

Change the camera view direction, bank, or zoom.  Called in response
to an arrow key press while in ARROW_MODE_VIEW.
====================================================================== */

static void change_view( int param, double d )
{
   switch ( param ) {
      case PARAM_LON:  view.lon = range( view.lon + d, TWOPI );  break;
      case PARAM_LAT:  view.lat += d;  break;
      case PARAM_BANK: view.bank += d; break;
      case PARAM_ZOOM: view.zoom *= d; break;
      default:
         break;
   }
   viewUpdate( &view );
   draw_stars();
}


/*
======================================================================
change_perr()

Change the systematic error (the shift) for the current constellation.
Called in response to an arrow key press while in ARROW_MODE_PTOLEMY.
====================================================================== */

static void change_perr( int param, float d )
{
   switch( param ) {
      case PARAM_LON:  errLon += d;  break;
      case PARAM_LAT:  errLat += d;  break;
      default:
         break;
   }
   draw_stars();
}


/*
======================================================================
change_epoch()

Change the epoch at which the stars are displayed.  This affects the
precession and proper motion of the field stars, and an offset may be
added to the Ptolemy stars.
====================================================================== */

static void change_epoch( void )
{
   int ok;
   
   ok = dlgEpoch( GTK_WINDOW( window ), &jd, &offsetLon, &offsetLat );
   if ( ok ) {
      posYale( star, jd );
      draw_stars();
   }
}


/*
======================================================================
change_cat()

Change the current catalog.  If the Info dialog is open, its contents
are updated.
====================================================================== */

static void change_cat( int c )
{
   if ( c > 0 && c <= 3 ) {
      pcon = pcat[ c - 1 ]->con;
      catname = pcat[ c - 1 ]->name;
      draw_stars();
      clickInfo( GTK_WINDOW( window ), star, &pcon[ curcon ], -1, -1,
         view.pradius, offsetLon, offsetLat );
   }
}


/*
======================================================================
replace_cat()

Load an Almagest catalog, replacing the current one with the new data.
====================================================================== */

static int replace_cat( void )
{
   char filename[ MAX_PATH_LEN ];
   PtolemyCatalog *newcat;
   int i, c = 0;
   
   /* find the index of the current catalog */
   
   for ( i = 0; i < 3; i++ )
      if ( pcon == pcat[ i ]->con ) {
         c = i + 1;
         break;
      }
   if ( !c ) return 0;

   /* display the file chooser */
   
   if ( !fileRequest( GTK_WINDOW( window ), NULL, filename, sizeof( filename ),
      FREQ_LOAD, "Load Catalog" ))
      return 0;

   /* try to load the new catalog */
   
   if ( !( newcat = readPCat( filename ))) {
      GtkWidget *msgdlg = gtk_message_dialog_new( NULL, 0, GTK_MESSAGE_ERROR,
         GTK_BUTTONS_CLOSE, "The new catalog couldn't be loaded." );
      gtk_dialog_run( GTK_DIALOG( msgdlg ));
      gtk_widget_destroy( msgdlg );
      return 0;
   }

   /* replace the current with the new and update the display */
   
   freePCat( pcat[ c - 1 ] );
   pcat[ c - 1 ] = newcat;
   change_cat( c );

   return 1;
}


/*
======================================================================
change_con()

Change the current constellation.  The current view is saved in the
view storage for the current constellation, then the constellation is
changed and the new view loaded and initialized.  The display is then
redrawn.
====================================================================== */

static void change_con( int d )
{
   put_conview( &conview[ curcon ] );
   curcon = d;
   if ( curcon < 0 ) curcon = NPCONS - 1;
   if ( curcon >= NPCONS ) curcon = 0;
   get_conview( &conview[ curcon ] );

   viewUpdate( &view );
   draw_stars();
}


/*
======================================================================
reset_conerr()

Reset the longitude and latitude error amounts for the Almagest stars,
and set the arrow mode.  If zero is TRUE, the errors are set to 0,
otherwise they're set to the defaults.
====================================================================== */

static void reset_conerr( int zero )
{
   ConViewDefaults cv;

   if ( zero )
      errLon = errLat = 0.0f;
   else {
      getConViewDefaults( &cv, curcon );
      errLon = cv.errLon;
      errLat = cv.errLat;
   }

   arrowmode = ARROW_MODE_PTOLEMY;
   draw_stars();
}


/*
======================================================================
reset_view()

Reset the view parameters to the current constellation defaults, and
set the arrow mode.
====================================================================== */

static void reset_view( void )
{
   ConViewDefaults cv;

   getConViewDefaults( &cv, curcon );
   view.lon  = cv.lon;
   view.lat  = cv.lat;
   view.bank = cv.bank;
   view.zoom = cv.zoom;

   arrowmode = ARROW_MODE_VIEW;
   viewUpdate( &view );
   draw_stars();
}


/*
======================================================================
change_vis()

Change layer visibility.  A layer is a drawing element:  the circles,
grids, ID lines, ID numbers, and Yale stars.  After turning one of
these on or off, the display needs to be redrawn.
====================================================================== */

static void change_vis( int vis )
{
   int bit;

   if ( vis < 1 || vis > 6 ) return;

   bit = 1 << ( vis - 1 );
   if ( layervis & bit )
      layervis &= ~bit;
   else
      layervis |= bit;

   draw_stars();
}


/*
======================================================================
change_color()

Allow the user to define an interface color.
====================================================================== */

static void change_color( int i )
{
   COLORREF c;

   c = getPrefColor( i );
   if ( selectColor( GTK_WINDOW( window ), &c )) {
      setPrefColor( i, c );
      draw_stars();
   }
}


/*
======================================================================
show_help()

Open the documentation in the default browser.

helpcmd[0] contains the filename of the browser program.  If this is
empty, the user is prompted for the filename, which will be saved in
the config file.  helpcmd[1] contains the fully qualified filename of
the help file, as read from the config file.  If this is empty,
findFile() is called to try to find it.

We use fork() and execlp() to start the browser.  fork() creates a
copy of the program's process at the point it's called.  The branch in
which fork() returns 0 is the child process, which we overlay with the
browser program using execlp().  If execlp() succeeds, execution of
the child process leaves here and enters the browser.  If it fails,
the new process immediately calls exit().
====================================================================== */

static void show_help( void )
{
   if ( !helpcmd[ 0 ] ) {
      char buf[ MAX_PATH_LEN ];
      if ( !fileRequest( GTK_WINDOW( window ), NULL, buf, sizeof( buf ),
         FREQ_LOAD, "Select a Browser" ))
         return;
      helpcmd[ 0 ] = malloc( strlen( buf ) + 1 );
      strcpy( helpcmd[ 0 ], buf );
   }

   if ( !helpcmd[ 1 ] ) {
      char buf[ MAX_PATH_LEN ];
      if ( !findFile( GTK_WINDOW( window ), "readme.html", buf, sizeof( buf )))
         return;
      helpcmd[ 1 ] = malloc( strlen( buf ) + 1 );
      strcpy( helpcmd[ 1 ], buf );
   }
      
   if ( fork() == 0 )
      if ( -1 == execlp( helpcmd[ 0 ], helpcmd[ 0 ], helpcmd[ 1 ], 0 ))
         exit( 0 );
}


/*
======================================================================
handle_arrowkey()

Called by handle_key() to deal with arrow key events.
====================================================================== */

static int handle_arrowkey( guint key, guint modifiers )
{
   int shift, control;
   float f;
   
   shift   = ( modifiers & GDK_SHIFT_MASK   ) != 0;
   control = ( modifiers & GDK_CONTROL_MASK ) != 0;

   if ( arrowmode == ARROW_MODE_VIEW ) {
      f = ( float ) DEGRAD( shift ? 10.0 : 1.0 );
      if ( control ) switch ( key ) {
         case GDK_Left:  change_view( PARAM_BANK, -f );  break;
         case GDK_Right: change_view( PARAM_BANK,  f );  break;
         default:  return 0;
      }
      else switch ( key ) {
         case GDK_Left:  change_view( PARAM_LON,  f );  break;
         case GDK_Right: change_view( PARAM_LON, -f );  break;
         case GDK_Up:    change_view( PARAM_LAT,  f );  break;
         case GDK_Down:  change_view( PARAM_LAT, -f );  break;
         default:  return 0;
      }
   }
   else if ( arrowmode == ARROW_MODE_PTOLEMY ) {
      f = ( float ) DEGRAD( shift ? 1.0 : 0.1 );
      switch ( key ) {
         case GDK_Left:  change_perr( PARAM_LON,  f );  break;
         case GDK_Right: change_perr( PARAM_LON, -f );  break;
         case GDK_Up:    change_perr( PARAM_LAT,  f );  break;
         case GDK_Down:  change_perr( PARAM_LAT, -f );  break;
         default:  return 0;
      }
   }

   return 0;
}


/*
======================================================================
handle_key()

Called by the ghandle_key() callback to handle key_press events.
====================================================================== */

static int handle_key( guint key, guint modifiers )
{
   if ( key > '0' && key < '7' ) {
      change_vis( key - '0' );
      return 0;
   }

   switch ( key ) {
      case GDK_Left:
      case GDK_Right:
      case GDK_Up:
      case GDK_Down:    handle_arrowkey( key, modifiers );    break;

      case GDK_Escape:  gtk_main_quit();           break;
      case GDK_Prior:   change_con( curcon - 1 );  break;
      case GDK_Next:    change_con( curcon + 1 );  break;
      case GDK_F1:      change_cat( 1 );           break;
      case GDK_F2:      change_cat( 2 );           break;
      case GDK_F3:      change_cat( 3 );           break;

      case ',':
      case '<':
         change_view( PARAM_ZOOM, 0.75 );
         break;
      case '.':
      case '>':
         change_view( PARAM_ZOOM, 1.0 / 0.75 );
         break;
      case 'v':
         change_arrowmode( ARROW_MODE_VIEW );
         break;
      case 'V':
         reset_view();
         break;
      case 's':
         if ( modifiers & GDK_CONTROL_MASK )
            reset_conerr( 1 );
         else
            change_arrowmode( ARROW_MODE_PTOLEMY );
         break;
      case 'S':
         reset_conerr( 0 );
         break;
      case 'c':
         replace_cat();
         break;
      case 'e':
         change_epoch();
         break;
      case 'p':
         pdf_stars();
         break;
      case 'h':
         show_help();
         break;
      case 'a':
         aboutDialog( GTK_WINDOW( window ), 0 );
         break;
      default: break;
   }
   return 0;
}


/*
======================================================================
ghandle_key()

The callback for key_press events.
====================================================================== */

static gint ghandle_key( GtkWidget *window, GdkEvent *event, gpointer data )
{
   GdkEventKey *e;
   
   if ( event == NULL ) return FALSE;
   if ( event->type != GDK_KEY_PRESS ) return FALSE;
   e = ( GdkEventKey * ) event;
   handle_key( e->keyval, e->state );
   return TRUE;
}


/*
======================================================================
handle_click()

One of the button_press_event callbacks, called to handle left mouse
clicks on the window.  Opens or updates the Info dialog.
====================================================================== */

static gint handle_click( GtkWidget *widget, GdkEvent *event, gpointer data )
{
   GdkEventButton *btnevent;
   
   if ( event == NULL ) return FALSE;
   if ( event->type != GDK_BUTTON_PRESS ) return FALSE;
   btnevent = ( GdkEventButton * ) event;
   if ( btnevent->button != 1 ) return FALSE;

   clickInfo( GTK_WINDOW( window ), star, &pcon[ curcon ],
      ( int ) btnevent->x, ( int ) btnevent->y, view.pradius, offsetLon,
      offsetLat );
   return TRUE;
}


/*
======================================================================
open_menu()

A button_press_event callback for right mouse clicks on the window.
The menu contents are synchronized with our current state (layer
visibility, arrow mode, catalog names) before calling gtk_menu_popup()
to open the menu at the position of the mouse click.
====================================================================== */

static gint open_menu( GtkWidget *window, GdkEvent *event, GtkWidget *wmenu )
{
   GtkMenu *menu;
   GdkEventButton *btnevent;
   int i, id;
   char buf[ 80 ];

   g_return_val_if_fail( wmenu != NULL, FALSE );
   g_return_val_if_fail( GTK_IS_MENU( wmenu ), FALSE );
   g_return_val_if_fail( event != NULL, FALSE );

   menu = GTK_MENU( wmenu );

   if ( event->type != GDK_BUTTON_PRESS ) return FALSE;
   btnevent = ( GdkEventButton * ) event;
   if ( btnevent->button != 3 ) return FALSE;

   for ( i = 0; i < 5; i++ )
      menuSetCheck( IDM_LVIS + i + 1, ( layervis & ( 1 << i )) != 0 );
   id = IDM_ARROWMODE + 1 + ( arrowmode == ARROW_MODE_PTOLEMY );
   menuSetCheck( id, 1 );
   for ( i = 0; i < 3; i++ ) {
      sprintf( buf, "F%d  %s", i + 1, pcat[ i ]->name );
      menuSetLabel( IDM_CAT + i + 1, buf );
   }

   gtk_menu_popup( menu, NULL, NULL, NULL, NULL, 
      btnevent->button, btnevent->time );
   return TRUE;
}


/*
======================================================================
handle_menu()

Called when the user selects a popup menu item.
====================================================================== */

static void handle_menu( GtkWidget *menuitem, int id )
{
   int major, minor;

   /* ignore radio item deselection */
   
   if ( GTK_IS_RADIO_MENU_ITEM( menuitem ))
      if ( !gtk_check_menu_item_get_active( GTK_CHECK_MENU_ITEM( menuitem )))
         return;

   minor = id % 100;
   major = id - minor;

   switch ( major ) {
      case IDM_CONS:
         change_con( minor - 1 );
         break;
      case IDM_LVIS:
         change_vis( minor );
         break;
      case IDM_ARROWMODE:
         switch ( minor ) {
            case 1:  handle_key( 'v', 0 );  break;
            case 2:  handle_key( 's', 0 );  break;
            case 3:  handle_key( 'V', 0 );  break;
            case 4:  handle_key( 'S', 0 );  break;
            case 5:  handle_key( 's', GDK_CONTROL_MASK );  break;
            default:  break;
         }
         break;
      case IDM_CAT:
         switch ( minor ) {
            case 1:
            case 2:
            case 3:
               change_cat( minor );
               break;
            case 4:  handle_key( 'c', 0 );  break;
            case 5:  handle_key( 'e', 0 );  break;
            default:  break;
         }
         break;
      case IDM_COLOR:
         change_color( minor - 1 );
         break;
      case IDM_HELP:
         handle_key( 'h', 0 );
         break;
      case IDM_PRINT:
         handle_key( 'p', 0 );
         break;
      case IDM_ABOUT:
         aboutDialog( GTK_WINDOW( window ), 0 );
         break;
      case IDM_QUIT:
         handle_key( GDK_Escape, 0 );
         break;
      default:
         break;
   }
}


/*
======================================================================
handle_size()

The configure_event callback, called when the window changes size.
We need to update our View structure, create a new pixmap, and create
a new drawing buffer.
====================================================================== */

static gboolean handle_size( GtkWidget *widget, GdkEventConfigure *event )
{
   if ( view.width == widget->allocation.width &&
        view.height == widget->allocation.height )
      return FALSE;
      
   if ( pixmap )
      g_object_unref( pixmap );

   view.width = widget->allocation.width;
   view.height = widget->allocation.height;

   pixmap = gdk_pixmap_new( widget->window, view.width, view.height, -1 );
   sizeDBuf( view.width, view.height );
   viewUpdate( &view );
   draw_stars();

   return TRUE;
}


/*
======================================================================
handle_expose()

Callback called when the window needs to be redrawn.
====================================================================== */

static gboolean handle_expose( GtkWidget *widget, GdkEventExpose *event )
{
   gdk_draw_drawable( widget->window, GC_FG, pixmap,
      event->area.x, event->area.y,
      event->area.x, event->area.y,
      event->area.width, event->area.height );
   return FALSE;
}


/*
======================================================================
handle_mousepos()

Callback called when the mouse cursor moves over the window.
====================================================================== */

static gboolean handle_mousepos( GtkWidget *widget, GdkEventMotion *event )
{
   int x, y;

   if ( event->is_hint ) {
      if ( !get_mousepos( &x, &y ))
         return TRUE;
   }
   else {
      x = event->x;
      y = event->y;
   }
    
   draw_mouse( x, y );

   return TRUE;
}


/*
======================================================================
destroy()

Callback called after the main window has been destroyed.
====================================================================== */

static void destroy( GtkWidget *widget, gpointer data )
{
   gtk_main_quit ();
}


/*
======================================================================
key_snoop()

Callback installed by gtk_key_snooper_install().

We want the user to be able to change catalogs while the Info dialog
is displayed.  The F1, F2, F3 keys would normally be received and
eaten by the child widgets on the dialog.  We intercept and redirect
those to the drawarea on the main window.
====================================================================== */

static gint key_snoop( GtkWidget *widget, GdkEventKey *event,
   gpointer func_data )
{
   if ( infoDialog ) {
      if ( widget == infoDialog || gtk_widget_is_ancestor( widget, infoDialog )) {
         if ( event->type == GDK_KEY_PRESS ) {
            switch ( event->keyval ) {
               case GDK_F1:
               case GDK_F2:
               case GDK_F3:
                  gtk_widget_event( drawarea, ( GdkEvent * ) event );
                  return TRUE;
            }
         }
      }
   }
   return FALSE;
}


/*
======================================================================
main()

Entry point.
====================================================================== */

int main( int argc, char *argv[] )
{
   GtkWidget *vbox;
   PangoFontDescription *fontdesc;
   char iconfile[ MAX_PATH_LEN ];
   int w, h;

   gtk_init( &argc, &argv );

   /* initialize from config file or defaults */

   loadConfig( cfgname, &w, &h, conview, helpcmd, &jd, &offsetLon, &offsetLat );

   /* load catalogs, quit if we can't */

   if ( !init_stars() ) {
      GtkWidget *msgdlg = gtk_message_dialog_new( NULL, 0, GTK_MESSAGE_ERROR,
         GTK_BUTTONS_CLOSE, "The catalog files couldn't be loaded." );
      gtk_dialog_run( GTK_DIALOG( msgdlg ));
      gtk_widget_destroy( msgdlg );
      return 0;
   }

   /* load the window icon */
   
   if ( findFile( NULL, "almstars.png", iconfile, sizeof( iconfile )))
      gtk_window_set_default_icon_from_file( iconfile, NULL );

   /* create the window and the drawing area */

   window = gtk_window_new( GTK_WINDOW_TOPLEVEL );
   gtk_window_set_title( GTK_WINDOW( window ), "Almagest Stars" );
   gtk_window_set_default_size( GTK_WINDOW( window ), w, h );
   drawarea = gtk_drawing_area_new();
   GTK_WIDGET_SET_FLAGS( drawarea, GTK_CAN_FOCUS );
   vbox = gtk_vbox_new( FALSE, 0 );
   gtk_container_add( GTK_CONTAINER( window ), vbox );
   gtk_widget_show( vbox );
   gtk_box_pack_start( GTK_BOX( vbox ), drawarea, TRUE, TRUE, 0 );
   gtk_widget_show( drawarea );

   /* get the font used by the status and mouse coordinate text */

   fontdesc = pango_font_description_from_string( "sans 10" );
   layout = gtk_widget_create_pango_layout( drawarea, NULL );
   pango_layout_set_font_description( layout, fontdesc );
   
   /* attach event handlers to the drawing area and window */
   
   gtk_widget_add_events( drawarea,
      GDK_BUTTON_PRESS_MASK   |
      GDK_KEY_PRESS_MASK      |
      GDK_EXPOSURE_MASK       |
      GDK_POINTER_MOTION_MASK |
      GDK_POINTER_MOTION_HINT_MASK );

   g_signal_connect( drawarea, "button-press-event",
      G_CALLBACK( handle_click ), NULL );
   g_signal_connect( drawarea, "key-press-event",
      G_CALLBACK( ghandle_key ), NULL );
   g_signal_connect( drawarea, "expose-event",
      G_CALLBACK( handle_expose ), NULL );
   g_signal_connect( drawarea, "configure-event",
      G_CALLBACK( handle_size ), NULL);
   g_signal_connect( drawarea, "motion-notify-event",
      G_CALLBACK( handle_mousepos ), NULL);

   g_signal_connect( window, "destroy",
      G_CALLBACK( destroy ), NULL );

   gtk_key_snooper_install( key_snoop, NULL );
   
   /* create and attach the popup menu */
   
   createMenu( drawarea, G_CALLBACK( handle_menu ), G_CALLBACK( open_menu ));

   /* show the window and the About dialog */

   gtk_widget_show( window );
   aboutDialog( GTK_WINDOW( window ), 5000 );

   /* message loop */

   gtk_main();
   
   /* done */

   put_conview( &conview[ curcon ] );
   saveConfig( cfgname, view.width, view.height, conview, helpcmd, jd,
      offsetLon, offsetLat );
   return 0;
}
