/*
======================================================================
yale.c

Ernie Wright  29 Aug 06

Functions for handling the bright star catalog.

   readYale()     read the bright star catalog
   posYale()      reduce positions to a given epoch
   clickYale()    find stars near a point on the main window
   HRtoIndex()    convert HR number to star array index

8 Jan 07    Rewrote properMotion() to use vecmat's left-handed system
            instead of Nash's right-handed one.  Corrected misplaced
            parentheses in the if() which prevented some candidates
            for the 3D calculation from being accepted.

----------------------------------------------------------------------
Copyright (c) 2007 by Ernie Wright

This file is part of the Linux version of Almagest Stars, which is
made available under the terms of version 2 of the GNU General Public
License (GPL) as published by the Free Software Foundation.  A copy of
the license should have been supplied to you in a file named COPYING
that accompanied this file.  A copy is also at http://www.gnu.org.
----------------------------------------------------------------------
====================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "astro.h"


/* convert from km/sec to parsecs/year */

#define PSECYR(x) ((x)*1.02272495e-6)

/* This allows us to look up indexes into the YaleStar array, given HR
   numbers.  Initialized by readYale() and used by HRtoIndex(). */

static short crossdir[ 9110 ];


/*
======================================================================
properMotion()

Calculate the change in right ascension and declination due to stellar
motion.  If the proper motion is large and the distance is small, the
new ra and dec are calculated the hard way, in 3D.  Otherwise they're
found with the simpler 2D calculation.  For most stars, the difference
between the 2D and 3D calculations is quite small.

The 3D calculation includes the side effect of altering the magnitude.
Based on the method described by Dave Nash,

   http://astronexus.com/node/40

The time t is in years, negative for times in the past.
====================================================================== */

static void properMotion( YaleStar *star, double *ra, double *dec,
   double t )
{
   Matrix3 M;
   double pos[ 3 ], v[ 3 ];

   if (( fabs( star->pmra ) > 0.01 || fabs( star->pmdec ) > 0.01 )
      && star->dist < 200.0 )
   {
      /* star position, parsecs */

      sphere_vec( star->ra, star->dec, pos );
      scalev( pos, star->dist );

      /* transverse motion, parsecs/yr */
      /* negate vta to get a left-handed vector */

      v[ 0 ] = PSECYR( -star->pmra  * star->dist * 4.740 );   /* vta */
      v[ 1 ] = PSECYR(  star->pmdec * star->dist * 4.740 );   /* vtd */
      v[ 2 ] = PSECYR(  star->rvel );

      /* transform to motion in our coordinate frame */

      identity3( M );
      mulrot3( M, rx_pitch( star->dec ), PITCH );
      mulrot3( M, ry_heading( star->ra ), HEADING );
      transform( v, M, v );

      /* add motion to position */

      pos[ 0 ] += v[ 0 ] * t;
      pos[ 1 ] += v[ 1 ] * t;
      pos[ 2 ] += v[ 2 ] * t;

      /* magnitude is affected by change in position */

      star->mag += ( float )( 5 * log10( vlen( pos ) / star->dist ));

      /* convert back to ra, dec */

      normalize( pos );
      vec_sphere( pos, ra, dec );
   }
   else
   {
      *ra  = star->ra  + DEGRAD( t * star->pmra  / ( 3600.0 * cos( star->dec )));
      *dec = star->dec + DEGRAD( t * star->pmdec / 3600.0 );
   }
}


/*
======================================================================
posYale()

Find the ecliptic coordinates of the Yale stars at the equinox and
epoch given by jd, and store them in the YaleStar.pos field as a LW
vector.
====================================================================== */

void posYale( YaleStar star[], double jd )
{
   Matrix3 M;
   double ra, dec, dy;
   int i;

   identity3( M );
   eq_ecl( JD_J2000, M );
   precess_ecl( jd, M );
   dy = jd_year( jd ) - 2000.0;

   for ( i = 0; i < 9096; i++ ) {
      properMotion( &star[ i ], &ra, &dec, dy );
      sphere_vec( ra, dec, star[ i ].pos );
      transform( star[ i ].pos, M, star[ i ].pos );
   }
}


/* Bayer letters and constellation abbreviations are stored in the
   Yale data file as indexes into these arrays. */

static char *greek[] = {
   "alp", "bet", "gam", "del", "eps", "zet", "eta", "the",
   "iot", "kap", "lam", "mu",  "nu",  "xi",  "omi", "pi",
   "rho", "sig", "tau", "ups", "phi", "chi", "psi", "ome"
};

static char *cons[] = {
   "And", "Ant", "Aps", "Aql", "Aqr", "Ara", "Ari", "Aur",
   "Boo", "CMa", "CMi", "CVn", "Cae", "Cam", "Cap", "Car",
   "Cas", "Cen", "Cep", "Cet", "Cha", "Cir", "Cnc", "Col",
   "Com", "CrA", "CrB", "Crt", "Cru", "Crv", "Cyg", "Del",
   "Dor", "Dra", "Equ", "Eri", "For", "Gem", "Gru", "Her",
   "Hor", "Hya", "Hyi", "Ind", "LMi", "Lac", "Leo", "Lep",
   "Lib", "Lup", "Lyn", "Lyr", "Men", "Mic", "Mon", "Mus",
   "Nor", "Oct", "Oph", "Ori", "Pav", "Peg", "Per", "Phe",
   "Pic", "PsA", "Psc", "Pup", "Pyx", "Ret", "Scl", "Sco",
   "Sct", "Ser", "Sex", "Sge", "Sgr", "Tau", "Tel", "TrA",
   "Tri", "Tuc", "UMa", "UMi", "Vel", "Vir", "Vol", "Vul"
};


/*
======================================================================
expand_id()

Expand the ID codes in the Yale data file into a human-readable
string.  The codes are single bytes representing the Flamsteed number,
Bayer letter, superscript, and constellation abbreviation.  The four-
byte hex sequence 08 01 01 31, for example, represents 8 alp1 Lib.
====================================================================== */

static void expand_id( unsigned char buf[], char name[] )
{
   int i = 0;

   if ( !buf[ 3 ] ) {
      name[ 0 ] = 0;
      return;
   }
   if ( buf[ 0 ] )
      i = sprintf( name, "%d ", ( int ) buf[ 0 ] );
   if ( buf[ 1 ] > 0 && buf[ 1 ] <= 24 )
      i += sprintf( name + i, "%s", greek[ buf[ 1 ] - 1 ] );
   if ( buf[ 2 ] > 0 && buf[ 2 ] < 10 )
      name[ i++ ] = '0' + buf[ 2 ];
   if ( buf[ 1 ] )
      name[ i++ ] = ' ';
   if ( buf[ 3 ] > 0 && buf[ 3 ] <= 88 )
      sprintf( name + i, "%s", cons[ buf[ 3 ] - 1 ] );
}


/*
======================================================================
verify_star()

Check that the values in a YaleStar are within valid ranges.
====================================================================== */

static int verify_star( YaleStar *star )
{
   if ( star->hr      < 1        || star->hr      > 9110    ) return 0;
   if ( star->hd      < 1        || star->hd      > 272150  ) return 0;
   if ( star->ra      < 0        || star->ra      > 6.2832f ) return 0;
   if ( star->dec     < -1.5708f || star->dec     > 1.5708f ) return 0;
   if ( star->dist    < 1.0f     || star->dist    > 1001.0f ) return 0;
   if ( star->pmra    < -6.0f    || star->pmra    > 6.0f    ) return 0;
   if ( star->pmdec   < -6.0f    || star->pmdec   > 6.0f    ) return 0;
   if ( star->rvel    < -400.0f  || star->rvel    > 400.0f  ) return 0;
   if ( star->mag     < -1.5f    || star->mag     > 8.0f    ) return 0;
   if ( star->bvindex < -0.5f    || star->bvindex > 6.0f    ) return 0;
   return 1;
}


/*
======================================================================
getI2()
getI4()
getF4()

Read little-endian binary values in a reasonably portable way.  We
aren't worried about negative I4's, since those are Henry Draper
catalog numbers.
====================================================================== */

static int getI2( FILE *fp )
{
   unsigned char p[ 2 ];
   int i;
   
   fread( p, 2, 1, fp );
   i = p[ 0 ] | p[ 1 ] << 8;
   if ( p[ 1 ] & 0x80 )
      i -= 65536;
   return i;
}


static int getI4( FILE *fp )
{
   unsigned char p[ 4 ];
   
   fread( p, 4, 1, fp );
   return p[ 0 ] | p[ 1 ] << 8 | p[ 2 ] << 16 | p[ 3 ] << 24;
}


static float getF4( FILE *fp )
{
   unsigned char p[ 4 ];
   unsigned long m;
   int s, e;
   
   fread( p, 4, 1, fp );
   s = ( p[ 3 ] & 0x80 ) ? -1 : 1;
   e = (( p[ 3 ] & 0x7F ) << 1 | ( p[ 2 ] & 0x80 ) >> 7 ) - 127;
   m = ( p[ 2 ] & 0x7F ) << 16 | p[ 1 ] << 8 | p[ 0 ];
   return ( float )( s * ldexp( 1.0 + m / 8388608.0, e ));
}


/*
======================================================================
readYale()

Read a file containing data on the 9096 stars in the Yale Bright Star
Catalog.  The file is binary and contains 9096 30-byte records with
the following format.

   hr    U2   Harvard Revised (Yale BSC) Catalog number
   hd    U4   Henry Draper Catalog number
   name {
      Flamsteed number   U1
      Bayer letter       U1
      superscript        U1
      constellation      U1
   }
   ra    F4   radians
   dec   F4   radians
   dist  U2   distance (parsecs) * 10
   pmra  I2   mas/yr
   pmdec I2   mas/yr
   rvel  I2   radial velocity (km/sec) * 10
   mag   I2   magnitude * 100
   bv    I2   B-V index * 100, or 999 for empty B-V data

The YaleStar array must have at least 9096 elements.  The function
fails, and returns 0, if the file can't be opened, if it's not the
expected size, or if any of the values are out of range.
====================================================================== */

int readYale( char *filename, YaleStar star[] )
{
   FILE *fp;
   int i, i2;
   unsigned char buf[ 4 ];

   fp = fopen( filename, "rb" );
   if ( !fp ) return 0;

   fseek( fp, 0, SEEK_END );
   if ( ftell( fp ) != 272880 ) {
      fclose( fp );
      return 0;
   }
   fseek( fp, 0, SEEK_SET );

   for ( i = 0; i < 9096; i++ ) {
      star[ i ].hr = getI2( fp );
      star[ i ].hd = getI4( fp );

      fread( buf, 4, 1, fp );
      expand_id( buf, star[ i ].name );

      star[ i ].ra    = getF4( fp );
      star[ i ].dec   = getF4( fp );
      star[ i ].dist  = getI2( fp ) /   10.0f;
      star[ i ].pmra  = getI2( fp ) / 1000.0f;
      star[ i ].pmdec = getI2( fp ) / 1000.0f;
      star[ i ].rvel  = getI2( fp ) /   10.0f;
      star[ i ].mag   = getI2( fp ) /  100.0f;

      i2 = getI2( fp );
      if ( i2 == 999 ) {
         star[ i ].cindex = 0;
         star[ i ].bvindex = 0.0f;
      }
      else {
         star[ i ].bvindex = i2 /  100.0f;
         star[ i ].cindex = indexFromBV( star[ i ].bvindex );
      }
      if ( !verify_star( &star[ i ] )) {
         fclose( fp );
         return 0;
      }
      crossdir[ star[ i ].hr ] = ( short ) i;
   }
   fclose( fp );
   return 1;
}


/* When the user clicks on the main window, we need to search for the
   closest Yale stars.  During the search, candidates are kept in
   sorted order in this array. */

static struct {
   float w;    /* sort on this, a combination of mag and distance */
   int   i;    /* star array index */
} sort[ 5 ];


/*
======================================================================
insert()

Insert a star in the candidate list in sorted order.
====================================================================== */

static void insert( int el, float w )
{
   int i, j;

   for ( i = 0; i < 5; i++ ) {
      if ( w < sort[ i ].w ) {
         for ( j = 4; j > i; j-- )
            sort[ j ] = sort[ j - 1 ];
         sort[ i ].w = w;
         sort[ i ].i = el;
         break;
      }
   }
}


/*
======================================================================
clear()

Clear the candidate list.
====================================================================== */

static void clear( void )
{
   int i;

   for ( i = 0; i < 5; i++ ) {
      sort[ i ].w = 1000.0f;
      sort[ i ].i = 0;
   }
}


/*
======================================================================
copy()

Copy the array indexes of the candidates to an array of int.
====================================================================== */

static void copy( int a[], int n )
{
   int i;

   for ( i = 0; i < n; i++ )
      a[ i ] = sort[ i ].i;
}


/*
======================================================================
clickYale()

Find stars within pradius pixels of a position on the main window.
Stars are sorted by a combination of magnitude and distance, so that
among candidates that are close to each other, the brightest is listed
first.  The array indexes of the found stars, up to 5, are copied to
found[], and the function returns the number of stars found.
====================================================================== */

int clickYale( YaleStar star[], int x, int y, double pradius, int found[] )
{
   double dx, dy, d;
   float w;
   int i, n = 0;

   clear();
   for ( i = 0; i < 9096; i++ ) {
      if ( !star[ i ].visible ) continue;
      dx = x - star[ i ].x;
      dy = y - star[ i ].y;
      d = sqrt( dx * dx + dy * dy );
      if ( d < pradius ) {
         w = ( float ) d * star[ i ].mag;
         insert( i, w );
         ++n;
      }
   }
   if ( n ) {
      if ( n > 5 ) n = 5;
      copy( found, n );
   }

   return n;
}


/*
======================================================================
HRtoIndex()

Convert an HR catalog number to an array index.
====================================================================== */

int HRtoIndex( int hr )
{
   if ( hr < 1 || hr > 9110 ) return -1;
   return crossdir[ hr ];
}
